# Daily Nitrate-Nitrite NOx and Chlorophyll-a Modeling

# import required packages
import os
import argparse
import pandas as pd
import numpy as np
from scipy.integrate import odeint
from loone.utils import loone_nchla_fns, load_config
from loone.data import Data as DClass


def _calculate_summer_month_loads(constit_loads_m: pd.DataFrame) -> tuple:
    """
    Calculate the summer month loads for NOx and Chla.

    Args:
        constit_loads_m (pd.DataFrame): The constituent loads DataFrame.

    Returns:
        tuple: A tuple containing arrays of summer month loads for NOx and Chla (StL and Cal).
    """
    Smr_Mnth_NOx_StL = []
    Smr_Mnth_NOx_Cal = []
    Smr_Mnth_Chla_StL = []
    Smr_Mnth_Chla_Cal = []
    for i in range(len(constit_loads_m.index)):
        if constit_loads_m['date'].iloc[i].month in [5, 6, 7, 8, 9, 10]:
            Smr_Mnth_NOx_StL.append(constit_loads_m['NO_Load_StL'].iloc[i])
            Smr_Mnth_Chla_StL.append(constit_loads_m['Chla_Load_StL'].iloc[i])
            Smr_Mnth_NOx_Cal.append(constit_loads_m['NO_Load_Cal'].iloc[i])
            Smr_Mnth_Chla_Cal.append(constit_loads_m['Chla_Load_Cal'].iloc[i])

    Smr_Mnth_NOx_StL_arr = np.asarray(Smr_Mnth_NOx_StL)
    Smr_Mnth_Chla_StL_arr = np.asarray(Smr_Mnth_Chla_StL)
    Smr_Mnth_NOx_Cal_arr = np.asarray(Smr_Mnth_NOx_Cal)
    Smr_Mnth_Chla_Cal_arr = np.asarray(Smr_Mnth_Chla_Cal)

    return Smr_Mnth_NOx_StL_arr, Smr_Mnth_Chla_StL_arr, Smr_Mnth_NOx_Cal_arr, Smr_Mnth_Chla_Cal_arr


def _prepare_constit_loads_df(inflows: pd.DataFrame, NO_Load_Cal: pd.Series, NO_Load_StL: pd.Series,
                              NO_Load_South: pd.Series, Chla_Load_Cal: pd.Series, Chla_Load_StL: pd.Series,
                              Chla_Load_South: pd.Series) -> tuple:
    """
    Prepare the constituent loads DataFrame and its monthly resampled version.

    Args:
        inflows (pd.DataFrame): The inflows DataFrame.
        NO_Load_Cal (pd.Series): The NO Load Cal series.
        NO_Load_StL (pd.Series): The NO Load StL series.
        NO_Load_South (pd.Series): The NO Load South series.
        Chla_Load_Cal (pd.Series): The Chla Load Cal series.
        Chla_Load_StL (pd.Series): The Chla Load StL series.
        Chla_Load_South (pd.Series): The Chla Load South series.

    Returns:
        tuple: A tuple containing the constituent loads DataFrame and its monthly resampled version.
    """
    constit_loads_df = pd.DataFrame(inflows['date'], columns=['date'])
    constit_loads_df['date'] = pd.to_datetime(constit_loads_df['date'])

    constit_loads_df['NO_Load_Cal'] = pd.to_numeric(NO_Load_Cal) / 1E9  # tons
    constit_loads_df['NO_Load_StL'] = pd.to_numeric(NO_Load_StL) / 1E9  # tons
    constit_loads_df['NO_Load_South'] = pd.to_numeric(NO_Load_South) / 1E9  # tons
    constit_loads_df['Chla_Load_Cal'] = pd.to_numeric(Chla_Load_Cal) / 1E6  # Kgs
    constit_loads_df['Chla_Load_StL'] = pd.to_numeric(Chla_Load_StL) / 1E6  # Kgs
    constit_loads_df['Chla_Load_South'] = pd.to_numeric(Chla_Load_South) / 1E6  # Kgs

    constit_loads_df = constit_loads_df.set_index('date')
    constit_loads_df.index = pd.to_datetime(constit_loads_df.index, unit='ns')

    constit_loads_m = constit_loads_df.resample('ME').sum()
    constit_loads_df = constit_loads_df.reset_index()
    constit_loads_m = constit_loads_m.reset_index()

    return constit_loads_df, constit_loads_m


def _update_nitro_model_output(nitro_model_output: pd.DataFrame, NO_N: pd.Series, NO_S: pd.Series, NO_MEAN: pd.Series, Sim_Chla: pd.Series, Sim_Chla_N: pd.Series, Sim_Chla_S: pd.Series, temperature: pd.Series) -> tuple:
    """
    Update the nitro model output DataFrame with the given series and resample it monthly.

    Args:
        nitro_model_output (pd.DataFrame): The nitro model output DataFrame.
        NO_N (pd.Series): The NO_N series.
        NO_S (pd.Series): The NO_S series.
        NO_MEAN (pd.Series): The NO_MEAN series.
        Sim_Chla (pd.Series): The Sim_Chla series.
        Sim_Chla_N (pd.Series): The Sim_Chla_N series.
        Sim_Chla_S (pd.Series): The Sim_Chla_S series.
        temperature (pd.Series): The temperature series.

    Returns:
        tuple: A tuple containing the updated nitro model output DataFrame and its monthly resampled version.
    """
    nitro_model_output['NO_N'] = pd.to_numeric(NO_N)
    nitro_model_output['NO_S'] = pd.to_numeric(NO_S)
    nitro_model_output['NO_M'] = pd.to_numeric(NO_MEAN)
    nitro_model_output['Sim_Chla'] = pd.to_numeric(Sim_Chla)
    nitro_model_output['Sim_Chla_N'] = pd.to_numeric(Sim_Chla_N)
    nitro_model_output['Sim_Chla_S'] = pd.to_numeric(Sim_Chla_S)
    nitro_model_output['Water Temp'] = pd.to_numeric(temperature)  # C

    nitro_model_output = nitro_model_output.set_index('date')
    nitro_model_output.index = pd.to_datetime(nitro_model_output.index, unit='ns')

    nitro_mod_out_m = nitro_model_output.resample('ME').mean()
    nitro_model_output = nitro_model_output.reset_index()
    nitro_mod_out_m = nitro_mod_out_m.reset_index()

    return nitro_model_output, nitro_mod_out_m


def _load_data(workspace: str, flow_path: str, forecast_mode: bool, photo_period_filename: str) -> dict:
    """
    Load the necessary data files from the workspace.

    Args:
        workspace (str): The path to the workspace directory.
        flow_path (str): The path to the flow data file.
        forecast_mode (bool): Whether to use forecast mode.
        photo_period_filename (str): The filename for the photo period data.

    Returns:
        dict: A dictionary containing the loaded data.
    """
    data = {}
    data['inflows'] = pd.read_csv(os.path.join(workspace, flow_path))
    data['temperature_data'] = pd.read_csv(os.path.join(workspace, 'Filled_WaterT.csv'))
    data['dissolved_oxygen'] = pd.read_csv(os.path.join(workspace, 'LO_DO_Clean_daily.csv'))
    data['radiation_data'] = pd.read_csv(os.path.join(workspace, 'LO_RADT_data.csv'))
    data['storage_data'] = pd.read_csv(os.path.join(workspace, 'Average_LO_Storage_3MLag.csv'))
    data['chlorophyll_a_north_data'] = pd.read_csv(os.path.join(workspace, 'N_Merged_Chla.csv'))  # microgram/L
    data['chlorophyll_a_south_data'] = pd.read_csv(os.path.join(workspace, 'S_Merged_Chla.csv'))  # microgram/L
    data['external_nitrate_loadings'] = pd.read_csv(os.path.join(workspace, 'LO_External_Loadings_NO.csv'))  # mg

    s65e_basename = 'water_quality_S65E_NITRATE+NITRITE-N_Interpolated_forecast.csv' if forecast_mode else 'water_quality_S65E_NITRATE+NITRITE-N_Interpolated.csv'
    data['s65e_nitrate_data'] = pd.read_csv(os.path.join(workspace, s65e_basename))  # mg/m3

    data['chlorophyll_a_loads_in'] = pd.read_csv(os.path.join(workspace, 'Chla_Loads_In.csv'))  # mg

    s65e_chlorophyll_a_basename = 'S65E_Chla_Merged_forecast.csv' if forecast_mode else 'S65E_Chla_Merged.csv'
    data['s65e_chlorophyll_a_data'] = pd.read_csv(os.path.join(workspace, s65e_chlorophyll_a_basename))  # mg/m3

    data['photo_period'] = pd.read_csv(os.path.join(workspace, f'{photo_period_filename}.csv'))
    data['lo_orthophosphate_north_data'] = pd.read_csv(os.path.join(workspace, 'N_OP.csv'))  # mg/m3
    data['lo_orthophosphate_south_data'] = pd.read_csv(os.path.join(workspace, 'S_OP.csv'))  # mg/m3
    data['lo_dissolved_inorganic_nitrogen_north_data'] = pd.read_csv(os.path.join(workspace, 'N_DIN.csv'))  # mg/m3
    data['lo_dissolved_inorganic_nitrogen_south_data'] = pd.read_csv(os.path.join(workspace, 'S_DIN.csv'))  # mg/m3

    return data


def _calculate_observed_values(lo_dissolved_inorganic_nitrogen_north_data: pd.DataFrame,
                               lo_dissolved_inorganic_nitrogen_south_data: pd.DataFrame,
                               chlorophyll_a_north_data: pd.DataFrame,
                               chlorophyll_a_south_data: pd.DataFrame,
                               lo_orthophosphate_north_data: pd.DataFrame,
                               lo_orthophosphate_south_data: pd.DataFrame) -> dict:
    """
    Calculate the observed values for various parameters.

    Args:
        lo_dissolved_inorganic_nitrogen_north_data (pd.DataFrame): The north dissolved inorganic nitrogen data.
        lo_dissolved_inorganic_nitrogen_south_data (pd.DataFrame): The south dissolved inorganic nitrogen data.
        chlorophyll_a_north_data (pd.DataFrame): The north chlorophyll-a data.
        chlorophyll_a_south_data (pd.DataFrame): The south chlorophyll-a data.
        lo_orthophosphate_north_data (pd.DataFrame): The north orthophosphate data.
        lo_orthophosphate_south_data (pd.DataFrame): The south orthophosphate data.

    Returns:
        dict: A dictionary containing the calculated observed values.
    """
    observed_values = {}

    observed_values['ammonium_north'] = lo_dissolved_inorganic_nitrogen_north_data['NH4'].astype(float)
    observed_values['ammonium_south'] = lo_dissolved_inorganic_nitrogen_south_data['NH4'].astype(float)
    observed_values['ammonium'] = (observed_values['ammonium_north'] + observed_values['ammonium_south']) / 2

    observed_values['nitrogen_oxide_north'] = lo_dissolved_inorganic_nitrogen_north_data['NO'].astype(float)
    observed_values['nitrogen_oxide_south'] = lo_dissolved_inorganic_nitrogen_south_data['NO'].astype(float)
    observed_values['nitrogen_oxide'] = (observed_values['nitrogen_oxide_north'] + observed_values['nitrogen_oxide_south']) / 2

    observed_values['chlorophyll_a_north'] = chlorophyll_a_north_data['Chla'].astype(float)
    observed_values['chlorophyll_a_south'] = chlorophyll_a_south_data['Chla'].astype(float)
    observed_values['chlorophyll_a'] = (observed_values['chlorophyll_a_north'] + observed_values['chlorophyll_a_south']) / 2

    observed_values['dissolved_inorganic_nitrogen_north'] = lo_dissolved_inorganic_nitrogen_north_data['DIN'].astype(float)
    observed_values['dissolved_inorganic_nitrogen_south'] = lo_dissolved_inorganic_nitrogen_south_data['DIN'].astype(float)
    observed_values['dissolved_inorganic_nitrogen'] = (observed_values['dissolved_inorganic_nitrogen_north'] + observed_values['dissolved_inorganic_nitrogen_south']) / 2

    observed_values['dissolved_inorganic_phosphorus_north'] = lo_orthophosphate_north_data['OP'].astype(float)
    observed_values['dissolved_inorganic_phosphorus_south'] = lo_orthophosphate_south_data['OP'].astype(float)
    observed_values['dissolved_inorganic_phosphorus'] = (observed_values['dissolved_inorganic_phosphorus_north'] + observed_values['dissolved_inorganic_phosphorus_south']) / 2

    return observed_values


def _get_light_attenuation(calibration_parameter_no: list, calibration_parameter_chla: list) -> dict:
    """
    Get light attenuation parameters.

    Args:
        calibration_parameter_no (list): The calibration parameters for NO.
        calibration_parameter_chla (list): The calibration parameters for Chla.

    Returns:
        dict: A dictionary containing the light attenuation parameters.
    """
    light_attenuation = {
        "Kw_NO": calibration_parameter_no[4],
        "Kw_Chla": calibration_parameter_chla[4],
        "Kc_NO": calibration_parameter_no[5],
        "Kc_Chla": calibration_parameter_chla[5],
    }
    return light_attenuation


def _get_light_limitation(calibration_parameter_no: list, calibration_parameter_chla: list) -> dict:
    """
    Get light limitation and inhibition coefficients.

    Args:
        calibration_parameter_no (list): The calibration parameters for NO.
        calibration_parameter_chla (list): The calibration parameters for Chla.

    Returns:
        dict: A dictionary containing the light limitation and inhibition coefficients.
    """
    light_limitation = {
        "K1_NO": calibration_parameter_no[6],
        "K1_Chla": calibration_parameter_chla[6],
        "K2_NO": calibration_parameter_no[7],
        "K2_Chla": calibration_parameter_chla[7],
    }
    return light_limitation


def _get_phytoplankton_growth(calibration_parameter_no: list, calibration_parameter_chla: list) -> dict:
    """
    Get maximum phytoplankton growth rate.

    Args:
        calibration_parameter_no (list): The calibration parameters for NO.
        calibration_parameter_chla (list): The calibration parameters for Chla.

    Returns:
        dict: A dictionary containing the maximum phytoplankton growth rate.
    """
    phytoplankton_growth = {
        "G_max_NO": calibration_parameter_no[8],
        "G_max_Chla": calibration_parameter_chla[8],
    }
    return phytoplankton_growth


def _get_half_saturation_coefficients(calibration_parameter_no: list, calibration_parameter_chla: list) -> dict:
    """
    Get half saturation coefficients.

    Args:
        calibration_parameter_no (list): The calibration parameters for NO.
        calibration_parameter_chla (list): The calibration parameters for Chla.

    Returns:
        dict: A dictionary containing the half saturation coefficients.
    """
    half_saturation = {
        "K_NH_NO": calibration_parameter_no[9],
        "K_NH_Chla": calibration_parameter_chla[9],
        "K_Nitr_NO": calibration_parameter_no[10],
        "K_Nitr_Chla": calibration_parameter_chla[10],
        "K_DIN_NO": calibration_parameter_no[9] + calibration_parameter_no[10],
        "K_DIN_Chla": calibration_parameter_chla[9] + calibration_parameter_chla[10],
        "KP_NO": calibration_parameter_no[11],
        "KP_Chla": calibration_parameter_chla[11],
        "K_TN_NO": calibration_parameter_no[12],
        "K_TN_Chla": calibration_parameter_chla[12],
        "YNOChla_NO": calibration_parameter_no[13],
        "YNOChla_Chla": calibration_parameter_chla[13],
    }
    return half_saturation


def _get_temperatures(calibration_parameter_no: list, calibration_parameter_chla: list) -> dict:
    """
    Get temperature parameters.

    Args:
        calibration_parameter_no (list): The calibration parameters for NO.
        calibration_parameter_chla (list): The calibration parameters for Chla.

    Returns:
        dict: A dictionary containing the temperature parameters.
    """
    temperatures = {
        "T_opt_NO": calibration_parameter_no[14],
        "T_min_NO": calibration_parameter_no[15],
        "T_max_NO": calibration_parameter_no[16],
        "T_opt_Chla": calibration_parameter_chla[14],
        "T_min_Chla": calibration_parameter_chla[15],
        "T_max_Chla": calibration_parameter_chla[16],
    }
    return temperatures


def _calculate_dissolved_oxygen(calibration_parameter_no: list, calibration_parameter_chla: list) -> dict:
    """
    Calculate dissolved oxygen parameters.

    Args:
        calibration_parameter_no (list): The calibration parameters for NO.
        calibration_parameter_chla (list): The calibration parameters for Chla.

    Returns:
        dict: A dictionary containing the dissolved oxygen parameters.
    """
    dissolved_oxygen = {
        "KDO_NO": calibration_parameter_no[17],
        "KDO_Chla": calibration_parameter_chla[17],
    }
    return dissolved_oxygen


def _calculate_sediment_release(calibration_parameter_no: list, calibration_parameter_chla: list) -> dict:
    """
    Calculate sediment release parameters.

    Args:
        calibration_parameter_no (list): The calibration parameters for NO.
        calibration_parameter_chla (list): The calibration parameters for Chla.

    Returns:
        dict: A dictionary containing the sediment release parameters.
    """
    sediment_release = {
        "S_NO_NO": calibration_parameter_no[18],
        "S_NO_Chla": calibration_parameter_chla[18],
    }
    return sediment_release


def _get_nitrification_parameters(calibration_parameter_no: list, calibration_parameter_chla: list) -> dict:
    """
    Get nitrification parameters for the detailed method.

    Args:
        calibration_parameter_no (list): The calibration parameters for NO.
        calibration_parameter_chla (list): The calibration parameters for Chla.

    Returns:
        dict: A dictionary containing the nitrification parameters.
    """
    nitrification_parameters = {
        "kni0_no": calibration_parameter_no[0],
        "kni0_chla": calibration_parameter_chla[0],
        "kni_no": calibration_parameter_no[1],
        "kni_chla": calibration_parameter_chla[1],
    }
    return nitrification_parameters


def _get_denitrification_parameters(calibration_parameter_no: list, calibration_parameter_chla: list) -> dict:
    """
    Get denitrification parameters for the detailed method.

    Args:
        calibration_parameter_no (list): The calibration parameters for NO.
        calibration_parameter_chla (list): The calibration parameters for Chla.

    Returns:
        dict: A dictionary containing the denitrification parameters.
    """
    denitrification_parameters = {
        "kden0_no": calibration_parameter_no[2],
        "kden0_chla": calibration_parameter_chla[2],
        "kden_no": calibration_parameter_no[3],
        "kden_chla": calibration_parameter_chla[3],
    }
    return denitrification_parameters


def _calculate_nutrient_limitation_factors(dissolved_inorganic_phosphorus_observed: float,
                                           dissolved_inorganic_nitrogen_observed: float,
                                           dissolved_inorganic_phosphorus_north_observed: float,
                                           dissolved_inorganic_phosphorus_south_observed: float,
                                           dissolved_inorganic_nitrogen_north_observed: float,
                                           dissolved_inorganic_nitrogen_south_observed: float,
                                           KP_NO: float, KP_Chla: float, K_DIN_NO: float, K_DIN_Chla: float) -> dict:
    """
    Calculate the nutrient limitation factors.

    Args:
        dissolved_inorganic_phosphorus_observed (float): The observed dissolved inorganic phosphorus.
        dissolved_inorganic_nitrogen_observed (float): The observed dissolved inorganic nitrogen.
        dissolved_inorganic_phosphorus_north_observed (float): The observed dissolved inorganic phosphorus in the north.
        dissolved_inorganic_phosphorus_south_observed (float): The observed dissolved inorganic phosphorus in the south.
        dissolved_inorganic_nitrogen_north_observed (float): The observed dissolved inorganic nitrogen in the north.
        dissolved_inorganic_nitrogen_south_observed (float): The observed dissolved inorganic nitrogen in the south.
        KP_NO (float): The KP_NO value.
        KP_Chla (float): The KP_Chla value.
        K_DIN_NO (float): The K_DIN_NO value.
        K_DIN_Chla (float): The K_DIN_Chla value.

    Returns:
        dict: A dictionary containing the nutrient limitation factors.
    """
    nutrient_limitation_factors = {
        "f_P_NO": dissolved_inorganic_phosphorus_observed / (KP_NO + dissolved_inorganic_phosphorus_observed),
        "f_P_Chla": dissolved_inorganic_phosphorus_observed / (KP_Chla + dissolved_inorganic_phosphorus_observed),
        "f_N_NO": dissolved_inorganic_nitrogen_observed / (K_DIN_NO + dissolved_inorganic_nitrogen_observed),
        "f_N_Chla": dissolved_inorganic_nitrogen_observed / (K_DIN_Chla + dissolved_inorganic_nitrogen_observed),
        "f_P_N_NO": dissolved_inorganic_phosphorus_north_observed / (KP_NO + dissolved_inorganic_phosphorus_north_observed),
        "f_P_N_Chla": dissolved_inorganic_phosphorus_north_observed / (KP_Chla + dissolved_inorganic_phosphorus_north_observed),
        "f_P_S_NO": dissolved_inorganic_phosphorus_south_observed / (KP_NO + dissolved_inorganic_phosphorus_south_observed),
        "f_P_S_Chla": dissolved_inorganic_phosphorus_south_observed / (KP_Chla + dissolved_inorganic_phosphorus_south_observed),
        "f_N_N_NO": dissolved_inorganic_nitrogen_north_observed / (K_DIN_NO + dissolved_inorganic_nitrogen_north_observed),
        "f_N_N_Chla": dissolved_inorganic_nitrogen_north_observed / (K_DIN_Chla + dissolved_inorganic_nitrogen_north_observed),
        "f_N_S_NO": dissolved_inorganic_nitrogen_south_observed / (K_DIN_NO + dissolved_inorganic_nitrogen_south_observed),
        "f_N_S_Chla": dissolved_inorganic_nitrogen_south_observed / (K_DIN_Chla + dissolved_inorganic_nitrogen_south_observed),
    }

    return nutrient_limitation_factors


def _calculate_inflows_and_outflows(i: int, storage_dev: list, q_i: list, q_o: list, external_nitrite_nitrate: list,
                                    s65e_nitrite_nitrate: list, external_chlorophyll_a: list,
                                    s65e_chlorophyll_a: list) -> tuple:
    """
    Calculate the inflows and outflows for the given index.

    Args:
        i (int): The current index.
        storage_dev (list): The storage deviation list.
        q_i (list): The inflow list.
        q_o (list): The outflow list.
        external_nitrite_nitrate (list): The external nitrite nitrate list.
        s65e_nitrite_nitrate (list): The S65E nitrite nitrate list.
        external_chlorophyll_a (list): The external chlorophyll-a list.
        s65e_chlorophyll_a (list): The S65E chlorophyll-a list.

    Returns:
        tuple: A tuple containing the calculated inflows and outflows.
    """
    if storage_dev[i] >= 0:
        Q_I_M = q_i[i] + storage_dev[i] * 1233.48  # m3/d
        Q_O_M = q_o[i]
        External_NO_M = external_nitrite_nitrate[i] + Q_I_M * s65e_nitrite_nitrate[i]
        External_Chla_M = external_chlorophyll_a[i] + Q_I_M * s65e_chlorophyll_a[i]
    else:
        Q_O_M = q_o[i] - storage_dev[i] * 1233.48  # m3/d
        Q_I_M = q_i[i]
        External_NO_M = external_nitrite_nitrate[i]
        External_Chla_M = external_chlorophyll_a[i]

    return Q_I_M, Q_O_M, External_NO_M, External_Chla_M


def _calculate_nitrification_rates(i: int, loone_nchla_fns: object, nit_denit_opt: str, total_nitrification_rate: float,
                                   kni0_no: float, kni_no: float, ammonium_north_observed: list,
                                   ammonium_south_observed: list, K_NH_NO: float, dissolved_oxygen: list, KDO_NO: float,
                                   fox_min: float, DO_Cr_n: float, DO_Opt_n: float, a: float, theta_ni: float,
                                   temperature: list) -> tuple:
    """
    Calculate the nitrification rates for the given index.

    Args:
        i (int): The current index.
        loone_nchla_fns (object): The LOONE NChla functions object.
        nit_denit_opt (str): The nitrification-denitrification option.
        total_nitrification_rate (float): The total nitrification rate.
        kni0_no (float): The kni0_no value.
        kni_no (float): The kni_no value.
        ammonium_north_observed (list): The observed ammonium in the north.
        ammonium_south_observed (list): The observed ammonium in the south.
        K_NH_NO (float): The K_NH_NO value.
        dissolved_oxygen (list): The dissolved oxygen list.
        KDO_NO (float): The KDO_NO value.
        fox_min (float): The fox_min value.
        DO_Cr_n (float): The DO_Cr_n value.
        DO_Opt_n (float): The DO_Opt_n value.
        a (float): The a value.
        theta_ni (float): The theta_ni value.
        temperature (list): The temperature list.

    Returns:
        tuple: A tuple containing the nitrification rates for the north and south regions.
    """
    Nit_R_N_NO = loone_nchla_fns.Nit_Rate('%s' % nit_denit_opt, total_nitrification_rate, kni0_no, kni_no,
                                          ammonium_north_observed[i], K_NH_NO, dissolved_oxygen[i], KDO_NO,
                                          fox_min, DO_Cr_n, DO_Opt_n, a)
    Nit_R_N_T_NO = Nit_R_N_NO * theta_ni ** (temperature[i] - 20)
    Nit_R_S_NO = loone_nchla_fns.Nit_Rate('%s' % nit_denit_opt, total_nitrification_rate, kni0_no, kni_no,
                                          ammonium_south_observed[i], K_NH_NO, dissolved_oxygen[i], KDO_NO,
                                          fox_min, DO_Cr_n, DO_Opt_n, a)
    Nit_R_S_T_NO = Nit_R_S_NO * theta_ni ** (temperature[i] - 20)

    return Nit_R_N_T_NO, Nit_R_S_T_NO


def _calculate_denitrification_rates(i: int, loone_nchla_fns: object, nit_denit_opt: str,
                                     total_denitrification_rate: float, kden0_no: float, kden_no: float, NO_N: list,
                                     NO_S: list, K_Nitr_NO: float, dissolved_oxygen: list, KDO_NO: float,
                                     DO_Cr_d: float, DO_Opt_d: float, theta_deni: float, temperature: list) -> tuple:
    """
    Calculate the denitrification rates for the given index.

    Args:
        i (int): The current index.
        loone_nchla_fns (object): The LOONE NChla functions object.
        nit_denit_opt (str): The nitrification-denitrification option.
        total_denitrification_rate (float): The total denitrification rate.
        kden0_no (float): The kden0_no value.
        kden_no (float): The kden_no value.
        NO_N (list): The NO in the north.
        NO_S (list): The NO in the south.
        K_Nitr_NO (float): The K_Nitr_NO value.
        dissolved_oxygen (list): The dissolved oxygen list.
        KDO_NO (float): The KDO_NO value.
        DO_Cr_d (float): The DO_Cr_d value.
        DO_Opt_d (float): The DO_Opt_d value.
        theta_deni (float): The theta_deni value.
        temperature (list): The temperature list.

    Returns:
        tuple: A tuple containing the denitrification rates for the north and south regions.
    """
    Denit_R_N_NO = loone_nchla_fns.Denit_Rate('%s' % nit_denit_opt, total_denitrification_rate, kden0_no,
                                              kden_no, NO_N[i], K_Nitr_NO, dissolved_oxygen[i], KDO_NO, DO_Cr_d,
                                              DO_Opt_d)
    Denit_R_N_T_NO = Denit_R_N_NO * theta_deni ** (temperature[i] - 20)
    Denit_R_S_NO = loone_nchla_fns.Denit_Rate('%s' % nit_denit_opt, total_denitrification_rate, kden0_no,
                                              kden_no, NO_S[i], K_Nitr_NO, dissolved_oxygen[i], KDO_NO, DO_Cr_d,
                                              DO_Opt_d)
    Denit_R_S_T_NO = Denit_R_S_NO * theta_deni ** (temperature[i] - 20)

    return Denit_R_N_T_NO, Denit_R_S_T_NO


def _calculate_nitrogen_oxide(i: int, loone_nchla_fns: object, temperature: list, T_opt_NO: float, T_min_NO: float,
                              T_max_NO: float, photo_period: pd.DataFrame, rad: list, Kw_NO: float, Kc_NO: float,
                              Sim_Chla: list, z: list, K1_NO: float, K2_NO: float, NO_N: list, t: np.ndarray,
                              external_nitrite_nitrate: list, atmospheric_deposition_north: float, Q_N2S: list,
                              Nit_R_N_T_NO: list, Denit_R_N_T_NO: list, ammonium_north_observed: list,
                              volume_north: list, G_max_NO: float, f_P_N_NO: list, f_N_N_NO: list, K_NH_NO: float,
                              K_TN_NO: float, YNOChla_NO: float, Sim_Chla_N: list, S_NO_NO: float, Area_N: float,
                              NO_Temp_Adj: list, atmospheric_deposition_south: float, Q_O_M: list, NO_S: list,
                              Nit_R_S_T_NO: list, Denit_R_S_T_NO: list, ammonium_south_observed: list,
                              volume_south: list, Sim_Chla_S: list, Area_S: float, NO_MEAN: list,
                              f_P_S_NO: list, f_N_S_NO: list) -> None:
    """
    Calculate the nitrogen oxide values for the given index.

    Args:
        i (int): The current index.
        loone_nchla_fns (object): The LOONE NChla functions object.
        temperature (list): The temperature list.
        T_opt_NO (float): The T_opt_NO value.
        T_min_NO (float): The T_min_NO value.
        T_max_NO (float): The T_max_NO value.
        photo_period (pd.DataFrame): The photo period DataFrame.
        rad (list): The radiation list.
        Kw_NO (float): The Kw_NO value.
        Kc_NO (float): The Kc_NO value.
        Sim_Chla (list): The simulated chlorophyll-a list.
        z (list): The depth list.
        K1_NO (float): The K1_NO value.
        K2_NO (float): The K2_NO value.
        NO_N (list): The NO in the north.
        t (np.ndarray): The time array.
        external_nitrite_nitrate (list): The external nitrite nitrate list.
        atmospheric_deposition_north (float): The atmospheric deposition in the north.
        Q_N2S (list): The Q_N2S list.
        Nit_R_N_T_NO (list): The nitrification rate in the north.
        Denit_R_N_T_NO (list): The denitrification rate in the north.
        ammonium_north_observed (list): The observed ammonium in the north.
        volume_north (list): The volume in the north.
        G_max_NO (float): The maximum growth rate for NO.
        f_P_N_NO (list): The phosphorus limitation factor for NO in the north.
        f_N_N_NO (list): The nitrogen limitation factor for NO in the north.
        K_NH_NO (float): The K_NH_NO value.
        K_TN_NO (float): The K_TN_NO value.
        YNOChla_NO (float): The YNOChla_NO value.
        Sim_Chla_N (list): The simulated chlorophyll-a in the north.
        S_NO_NO (float): The S_NO_NO value.
        Area_N (float): The area in the north.
        NO_Temp_Adj (list): The temperature adjustment for NO.
        atmospheric_deposition_south (float): The atmospheric deposition in the south.
        Q_O_M (list): The Q_O_M list.
        NO_S (list): The NO in the south.
        Nit_R_S_T_NO (list): The nitrification rate in the south.
        Denit_R_S_T_NO (list): The denitrification rate in the south.
        ammonium_south_observed (list): The observed ammonium in the south.
        volume_south (list): The volume in the south.
        Sim_Chla_S (list): The simulated chlorophyll-a in the south.
        Area_S (float): The area in the south.
        NO_MEAN (list): The mean NO.
        f_P_S_NO (list): The phosphorus limitation factor for NO in the south.
        f_N_S_NO (list): The nitrogen limitation factor for NO in the south.

    Returns:
        None
    """
    fT_NO = loone_nchla_fns.f_T_alt1(temperature[i], T_opt_NO, T_min_NO, T_max_NO)
    fL_NO = loone_nchla_fns.f_L_alt1(photo_period['Data'].iloc[i], rad[i], Kw_NO, Kc_NO, Sim_Chla[i], z[i],
                                     K1_NO, K2_NO)

    DfEq_Res_N = odeint(loone_nchla_fns.NOx_N_DiffEq, NO_N[i], t,
                        args=(external_nitrite_nitrate[i], atmospheric_deposition_north, Q_N2S[i], Nit_R_N_T_NO[i],
                              Denit_R_N_T_NO[i], ammonium_north_observed[i], volume_north[i], G_max_NO, fT_NO,
                              fL_NO, f_P_N_NO[i], f_N_N_NO[i], K_NH_NO, K_TN_NO, YNOChla_NO, Sim_Chla_N[i],
                              S_NO_NO, Area_N, NO_Temp_Adj[i], ))
    NO_N[i + 1] = DfEq_Res_N[:, 0][1]
    DfEq_Res_S = odeint(loone_nchla_fns.NOx_S_DiffEq, NO_S[i], t,
                        args=(atmospheric_deposition_south, Q_N2S[i], Q_O_M[i], NO_N[i], Nit_R_S_T_NO[i],
                              Denit_R_S_T_NO[i], ammonium_south_observed[i], volume_south[i], G_max_NO, fT_NO,
                              fL_NO, f_P_S_NO[i], f_N_S_NO[i], K_NH_NO, K_TN_NO, YNOChla_NO, Sim_Chla_S[i],
                              S_NO_NO, Area_S, NO_Temp_Adj[i], ))
    NO_S[i + 1] = DfEq_Res_S[:, 0][1]
    NO_MEAN[i + 1] = (NO_N[i + 1] + NO_S[i + 1]) * 0.5


def _calculate_no_loads(i: int, NO_S: list, s77_outflow: list, s308_outflow: list,
                        total_regional_outflow_south: list) -> tuple:
    """
    Calculate the NO loads for the given index.

    Args:
        i (int): The current index.
        NO_S (list): The NO in the south.
        s77_outflow (list): The S77 outflow list.
        s308_outflow (list): The S308 outflow list.
        total_regional_outflow_south (list): The total regional outflow in the south.

    Returns:
        tuple: A tuple containing the NO loads for Cal, StL, and South.
    """
    NO_Load_Cal = s77_outflow[i] * NO_S[i]  # mg/d P
    NO_Load_StL = s308_outflow[i] * NO_S[i]  # mg/d P
    NO_Load_South = total_regional_outflow_south[i] * 1233.48 * NO_S[i]  # mg/d P

    return NO_Load_Cal, NO_Load_StL, NO_Load_South


def _calculate_chlorophyll_a(i: int, loone_nchla_fns: object, temperature: list, T_opt_Chla: float, T_min_Chla: float,
                             T_max_Chla: float, nitro_model_output: pd.DataFrame, photo_period: pd.DataFrame,
                             rad: list, Kw_Chla: float, Kc_Chla: float, Sim_Chla: list, z: list, K1_Chla: float,
                             K2_Chla: float, External_Chla_M: list, Q_N2S: list, Sim_Chla_N: list, vc: float,
                             K_m_T: list, K_r_T: list, G_max_Chla: float, f_P_N_Chla: list, f_N_N_Chla: list,
                             volume_north: list, Q_O_M: list, Sim_Chla_S: list, f_P_S_Chla: list, f_N_S_Chla: list,
                             volume_south: list) -> None:
    """
    Calculate the chlorophyll-a values for the given index.

    Args:
        i (int): The current index.
        loone_nchla_fns (object): The LOONE NChla functions object.
        temperature (list): The temperature list.
        T_opt_Chla (float): The T_opt_Chla value.
        T_min_Chla (float): The T_min_Chla value.
        T_max_Chla (float): The T_max_Chla value.
        nitro_model_output (pd.DataFrame): The nitro model output DataFrame.
        photo_period (pd.DataFrame): The photo period DataFrame.
        rad (list): The radiation list.
        Kw_Chla (float): The Kw_Chla value.
        Kc_Chla (float): The Kc_Chla value.
        Sim_Chla (list): The simulated chlorophyll-a list.
        z (list): The depth list.
        K1_Chla (float): The K1_Chla value.
        K2_Chla (float): The K2_Chla value.
        External_Chla_M (list): The external chlorophyll-a list.
        Q_N2S (list): The Q_N2S list.
        Sim_Chla_N (list): The simulated chlorophyll-a in the north.
        vc (float): The vc value.
        K_m_T (list): The K_m_T value.
        K_r_T (list): The K_r_T value.
        G_max_Chla (float): The maximum growth rate for chlorophyll-a.
        f_P_N_Chla (list): The phosphorus limitation factor for chlorophyll-a in the north.
        f_N_N_Chla (list): The nitrogen limitation factor for chlorophyll-a in the north.
        volume_north (list): The volume in the north.
        Q_O_M (list): The Q_O_M list.
        Sim_Chla_S (list): The simulated chlorophyll-a in the south.
        f_P_S_Chla (list): The phosphorus limitation factor for chlorophyll-a in the south.
        f_N_S_Chla (list): The nitrogen limitation factor for chlorophyll-a in the south.
        volume_south (list): The volume in the south.

    Returns:
        None
    """
    fT_Chla = loone_nchla_fns.f_T__Chla_alt1(temperature[i], T_opt_Chla, T_min_Chla, T_max_Chla, nitro_model_output['date'].iloc[i].month)
    fL_Chla = loone_nchla_fns.f_L_alt1(photo_period['Data'].iloc[i], rad[i], Kw_Chla, Kc_Chla, Sim_Chla[i], z[i], K1_Chla, K2_Chla) * 1.2 if nitro_model_output['date'].iloc[i].month in (6, 7, 8, 9, 10) else loone_nchla_fns.f_L_alt1(photo_period['Data'].iloc[i], rad[i], Kw_Chla, Kc_Chla, Sim_Chla[i], z[i], K1_Chla, K2_Chla) * 1

    Sim_Chla_N[i + 1] = loone_nchla_fns.Chla_N_alt1(External_Chla_M[i], Q_N2S[i], Sim_Chla_N[i], vc, z[i], K_m_T[i], K_r_T[i], 0, G_max_Chla, fT_Chla, fL_Chla, f_P_N_Chla[i], f_N_N_Chla[i], volume_north[i])
    Sim_Chla_S[i + 1] = loone_nchla_fns.Chla_S_alt1(Q_N2S[i], Q_O_M[i], Sim_Chla_N[i], Sim_Chla_S[i], vc, z[i], K_m_T[i], K_r_T[i], 0, G_max_Chla, fT_Chla, fL_Chla, f_P_S_Chla[i], f_N_S_Chla[i], volume_south[i])
    Sim_Chla[i + 1] = (Sim_Chla_N[i + 1] + Sim_Chla_S[i + 1]) / 2


def _calculate_chla_loads(i: int, Sim_Chla_S: list, s77_outflow: list, s308_outflow: list,
                          total_regional_outflow_south: list) -> tuple:
    """
    Calculate the chlorophyll-a loads for the given index.

    Args:
        i (int): The current index.
        Sim_Chla_S (list): The simulated chlorophyll-a in the south.
        s77_outflow (list): The S77 outflow list.
        s308_outflow (list): The S308 outflow list.
        total_regional_outflow_south (list): The total regional outflow in the south.

    Returns:
        tuple: A tuple containing the chlorophyll-a loads for Cal, StL, and South.
    """
    Chla_Load_Cal = s77_outflow[i] * Sim_Chla_S[i]  # mg/d P
    Chla_Load_StL = s308_outflow[i] * Sim_Chla_S[i]  # mg/d P
    Chla_Load_South = total_regional_outflow_south[i] * 1233.48 * Sim_Chla_S[i]  # mg/d P

    return Chla_Load_Cal, Chla_Load_StL, Chla_Load_South


def LOONE_WQ(workspace: str, photo_period_filename: str = 'PhotoPeriod', forecast_mode: bool = False) -> list:
    """Daily Nitrate-Nitrite NOx and Chlorophyll-a Modeling

    Args:
        workspace (str): The working directory path.
        photo_period_filename (str): The filename of the photo period data. Default is 'PhotoPeriod'.
        forecast_mode (bool): The forecast mode flag.

    Returns:
        list: The results of the simulation.
    """
    # Read in the config file
    config = load_config(workspace)

    # Initialize the Data object
    data = DClass(workspace)

    # Read Required Data
    flow_path = os.path.join(workspace, 'LO_Inflows_BK_forecast.csv' if forecast_mode else 'LO_Inflows_BK.csv')

    data_dict = _load_data(workspace, flow_path, forecast_mode, photo_period_filename)

    inflows = data_dict['inflows']
    temperature_data = data_dict['temperature_data']
    dissolved_oxygen = data_dict['dissolved_oxygen']
    radiation_data = data_dict['radiation_data']
    storage_data = data_dict['storage_data']
    chlorophyll_a_north_data = data_dict['chlorophyll_a_north_data']
    chlorophyll_a_south_data = data_dict['chlorophyll_a_south_data']
    external_nitrate_loadings = data_dict['external_nitrate_loadings']
    s65e_nitrate_data = data_dict['s65e_nitrate_data']
    chlorophyll_a_loads_in = data_dict['chlorophyll_a_loads_in']
    s65e_chlorophyll_a_data = data_dict['s65e_chlorophyll_a_data']
    photo_period = data_dict['photo_period']
    lo_orthophosphate_north_data = data_dict['lo_orthophosphate_north_data']
    lo_orthophosphate_south_data = data_dict['lo_orthophosphate_south_data']
    lo_dissolved_inorganic_nitrogen_north_data = data_dict['lo_dissolved_inorganic_nitrogen_north_data']
    lo_dissolved_inorganic_nitrogen_south_data = data_dict['lo_dissolved_inorganic_nitrogen_south_data']

    date_start = inflows['date'].iloc[0]

    rad = radiation_data['Mean_RADT'].astype(float) * 4.6 * 1000

    temperature = temperature_data['Water_T'].astype(float)
    dissolved_oxygen = dissolved_oxygen['Mean_DO'].astype(float)
    external_nitrite_nitrate = external_nitrate_loadings['External_NO_Ld_mg'].astype(float)  # mg
    s65e_nitrite_nitrate = (s65e_nitrate_data[s65e_nitrate_data['date'] >= date_start]['Data'] * 1000).astype(float).tolist()  # mg/m3

    external_chlorophyll_a = chlorophyll_a_loads_in['Chla_Loads'].astype(float) * 3  # mg
    s65e_chlorophyll_a = s65e_chlorophyll_a_data[s65e_chlorophyll_a_data['date'] >= date_start]['Data'].astype(float).tolist()

    # N-S Procedure
    N_Per = 0.43
    S_Per = 0.57

    storage_dev = data.Storage_dev_df['DS_dev'].astype(float)  # acft
    q_i = inflows['Inflows_cmd'].astype(float)  # m3

    # Simulated Q
    # S77_Q = LOONE_Q_Outputs['S77_Q'] * 0.0283168 * 86400  # cfs to cubic meters per day
    # S308_Q = LOONE_Q_Outputs['S308_Q'] * 0.0283168 * 86400  # cfs to cubic meters per day
    # TotRegSo = LOONE_Q_Outputs['TotRegSo']  # acft/day
    # TotRegEW = LOONE_Q_Outputs['TotRegEW']  # acft/day

    # Simulated Stage and Storage
    # Stage = LOONE_Q_Outputs['Stage'] * 0.3048  # ft to m
    # volume = LOONE_Q_Outputs['Storage'] * 1233.48  # acft to m3

    # Observed S77 S308 South
    outflows_observed = pd.read_csv(os.path.join(workspace, 'Flow_df_3MLag.csv'))
    s77_outflow = outflows_observed['S77_Out']
    s308_outflow = outflows_observed['S308_Out']
    total_regional_outflow_south = outflows_observed[['S351_Out', 'S354_Out', 'S352_Out', 'L8_Out']].sum(axis=1) / 1233.48    # m3/day to acft

    # Observed Stage and Storage
    stage = storage_data['Stage_ft'].astype(float) * 0.3048  # m
    volume = storage_data['Storage_cmd'].astype(float)  # m3

    volume_north = volume * N_Per
    volume_south = volume * S_Per

    q_o = s77_outflow + s308_outflow + total_regional_outflow_south * 1233.48  # cmd

    observed_values = _calculate_observed_values(lo_dissolved_inorganic_nitrogen_north_data, lo_dissolved_inorganic_nitrogen_south_data, chlorophyll_a_north_data, chlorophyll_a_south_data, lo_orthophosphate_north_data, lo_orthophosphate_south_data)

    ammonium_north_observed = observed_values['ammonium_north']
    ammonium_south_observed = observed_values['ammonium_south']
    ammonium_observed = observed_values['ammonium']
    nitrogen_oxide_north_observed = observed_values['nitrogen_oxide_north']
    nitrogen_oxide_south_observed = observed_values['nitrogen_oxide_south']
    nitrogen_oxide_observed = observed_values['nitrogen_oxide']
    chlorophyll_a_north = observed_values['chlorophyll_a_north']
    chlorophyll_a_south = observed_values['chlorophyll_a_south']
    chlorophyll_a = observed_values['chlorophyll_a']
    dissolved_inorganic_nitrogen_north_observed = observed_values['dissolved_inorganic_nitrogen_north']
    dissolved_inorganic_nitrogen_south_observed = observed_values['dissolved_inorganic_nitrogen_south']
    dissolved_inorganic_nitrogen_observed = observed_values['dissolved_inorganic_nitrogen']
    dissolved_inorganic_phosphorus_north_observed = observed_values['dissolved_inorganic_phosphorus_north']
    dissolved_inorganic_phosphorus_south_observed = observed_values['dissolved_inorganic_phosphorus_south']
    dissolved_inorganic_phosphorus_observed = observed_values['dissolved_inorganic_phosphorus']

    nit_denit_opt = 'Opt1'

    # NO Atmospheric Deposition
    atmospheric_deposition = (714.6 * 1000 * 1000)  # mg/day
    atmospheric_deposition_north = N_Per * atmospheric_deposition
    atmospheric_deposition_south = S_Per * atmospheric_deposition

    # Calibration parameters
    calibration_parameter_no = data.Cal_Par['Par_NO']
    calibration_parameter_chla = data.Cal_Par['Par_Chla']

    # nitrification
    # Either a simple method where Nit_R_T = nitrification rate in water column (Tot_nit_R) * Temp Coeff for nitrification * (T-20)
    # Or a more detailed method R = k0 + k * fam * fox
    # two options for fam and fox
    # Option1
    # fam = (Cam/Ks+Cam) and fox = (Cox/Ksox+Cox)
    # Option 2
    # Fam = Cam and fox = (1-foxmin) * (Cox-Coxc/Coxo-Coxc) ^10a + foxmin
    # The nitrification rate for the simple method
    total_nitrification_rate = 0.3
    # nitrification parameters for the detailed method
    nitrification_parameters = _get_nitrification_parameters(calibration_parameter_no, calibration_parameter_chla)
    kni0_no = nitrification_parameters["kni0_no"]
    kni0_chla = nitrification_parameters["kni0_chla"]
    kni_no = nitrification_parameters["kni_no"]
    kni_chla = nitrification_parameters["kni_chla"]

    theta_ni = 1.06

    # Denitrification
    # Either a simple method where Denit_R_T = denitrification rate in water column (Tot_denit_R) * Temp Coeff for denitrification * (T-20)
    # Or a more detailed method R = k0 + k * fni * fox
    # two options for fam and fox
    # Option1
    # fni = (Cni/Ks+Cni) and fox = 1- (Cox/Ksox+Cox)
    # Option 2
    # Fni = Cni and fox = (Coxc-Cox/Coxc-Coxo)
    # The denitrification rate for the simple method
    total_denitrification_rate = 0.1
    # The denitrification rate for the simple method
    denitrification_parameters = _get_denitrification_parameters(calibration_parameter_no, calibration_parameter_chla)
    kden0_no = denitrification_parameters["kden0_no"]
    kden0_chla = denitrification_parameters["kden0_chla"]
    kden_no = denitrification_parameters["kden_no"]
    kden_chla = denitrification_parameters["kden_chla"]

    theta_deni = 1.06

    # Light Attenuation due to water Kw and algae Kc
    light_attenuation = _get_light_attenuation(calibration_parameter_no, calibration_parameter_chla)
    Kw_NO = light_attenuation["Kw_NO"]
    Kw_Chla = light_attenuation["Kw_Chla"]
    Kc_NO = light_attenuation["Kc_NO"]
    Kc_Chla = light_attenuation["Kc_Chla"]

    # Light limitation and inhibition coefficients K1 and K2
    light_limitation = _get_light_limitation(calibration_parameter_no, calibration_parameter_chla)
    K1_NO = light_limitation["K1_NO"]
    K1_Chla = light_limitation["K1_Chla"]
    K2_NO = light_limitation["K2_NO"]
    K2_Chla = light_limitation["K2_Chla"]

    # Maximum Phytoplankton growth rate
    phytoplankton_growth = _get_phytoplankton_growth(calibration_parameter_no, calibration_parameter_chla)
    G_max_NO = phytoplankton_growth["G_max_NO"]
    G_max_Chla = phytoplankton_growth["G_max_Chla"]

    # Half Saturation Coefficients
    half_saturation = _get_half_saturation_coefficients(calibration_parameter_no, calibration_parameter_chla)
    K_NH_NO = half_saturation["K_NH_NO"]
    K_NH_Chla = half_saturation["K_NH_Chla"]
    K_Nitr_NO = half_saturation["K_Nitr_NO"]
    K_Nitr_Chla = half_saturation["K_Nitr_Chla"]
    K_DIN_NO = half_saturation["K_DIN_NO"]
    K_DIN_Chla = half_saturation["K_DIN_Chla"]
    KP_NO = half_saturation["KP_NO"]
    KP_Chla = half_saturation["KP_Chla"]
    K_TN_NO = half_saturation["K_TN_NO"]
    K_TN_Chla = half_saturation["K_TN_Chla"]
    YNOChla_NO = half_saturation["YNOChla_NO"]
    YNOChla_Chla = half_saturation["YNOChla_Chla"]

    # Temperatures
    temperatures = _get_temperatures(calibration_parameter_no, calibration_parameter_chla)
    T_opt_NO = temperatures["T_opt_NO"]
    T_min_NO = temperatures["T_min_NO"]
    T_max_NO = temperatures["T_max_NO"]
    T_opt_Chla = temperatures["T_opt_Chla"]
    T_min_Chla = temperatures["T_min_Chla"]
    T_max_Chla = temperatures["T_max_Chla"]

    # Dissolved Oxygen
    dissolved_oxygen_params = _calculate_dissolved_oxygen(calibration_parameter_no, calibration_parameter_chla)
    KDO_NO = dissolved_oxygen_params["KDO_NO"]
    KDO_Chla = dissolved_oxygen_params["KDO_Chla"]

    # Sediment Release
    sediment_release = _calculate_sediment_release(calibration_parameter_no, calibration_parameter_chla)
    S_NO_NO = sediment_release["S_NO_NO"]
    S_NO_Chla = sediment_release["S_NO_Chla"]

    Theta_NO = 1.06
    NO_Temp_Adj = Theta_NO ** (temperature - 20)
    Area = 1730 * 1E6  # m2
    Area_N = Area * N_Per
    Area_S = Area * S_Per

    X = len(inflows.index)

    # Nitrogen and Phosphorus Limiting
    nutrient_limitation_factors = _calculate_nutrient_limitation_factors(dissolved_inorganic_phosphorus_observed, dissolved_inorganic_nitrogen_observed, dissolved_inorganic_phosphorus_north_observed, dissolved_inorganic_phosphorus_south_observed, dissolved_inorganic_nitrogen_north_observed, dissolved_inorganic_nitrogen_south_observed, KP_NO, KP_Chla, K_DIN_NO, K_DIN_Chla)

    f_P_NO = nutrient_limitation_factors["f_P_NO"]
    f_P_Chla = nutrient_limitation_factors["f_P_Chla"]
    f_N_NO = nutrient_limitation_factors["f_N_NO"]
    f_N_Chla = nutrient_limitation_factors["f_N_Chla"]
    f_P_N_NO = nutrient_limitation_factors["f_P_N_NO"]
    f_P_N_Chla = nutrient_limitation_factors["f_P_N_Chla"]
    f_P_S_NO = nutrient_limitation_factors["f_P_S_NO"]
    f_P_S_Chla = nutrient_limitation_factors["f_P_S_Chla"]
    f_N_N_NO = nutrient_limitation_factors["f_N_N_NO"]
    f_N_N_Chla = nutrient_limitation_factors["f_N_N_Chla"]
    f_N_S_NO = nutrient_limitation_factors["f_N_S_NO"]
    f_N_S_Chla = nutrient_limitation_factors["f_N_S_Chla"]

    Q_I_M = np.zeros(X, dtype=object)
    Q_O_M = np.zeros(X, dtype=object)
    Q_N2S = np.zeros(X, dtype=object)
    External_NO_M = np.zeros(X, dtype=object)
    External_Chla_M = np.zeros(X, dtype=object)

    Nit_R_N_NO = np.zeros(X, dtype=object)
    Nit_R_N_T_NO = np.zeros(X, dtype=object)
    Nit_R_S_NO = np.zeros(X, dtype=object)
    Nit_R_S_T_NO = np.zeros(X, dtype=object)
    Denit_R_N_NO = np.zeros(X, dtype=object)
    Denit_R_N_T_NO = np.zeros(X, dtype=object)
    Denit_R_S_NO = np.zeros(X, dtype=object)
    Denit_R_S_T_NO = np.zeros(X, dtype=object)

    fT_NO = np.zeros(X, dtype=object)
    fL_NO = np.zeros(X, dtype=object)

    fT_Chla = np.zeros(X, dtype=object)
    fL_Chla = np.zeros(X, dtype=object)

    NO_N = np.zeros(X, dtype=object)
    NO_S = np.zeros(X, dtype=object)
    NO_MEAN = np.zeros(X, dtype=object)

    NO_Load_Cal = np.zeros(X, dtype=object)
    NO_Load_StL = np.zeros(X, dtype=object)
    NO_Load_South = np.zeros(X, dtype=object)

    Sim_Chla_N = np.zeros(X, dtype=object)
    Sim_Chla_S = np.zeros(X, dtype=object)
    Sim_Chla = np.zeros(X, dtype=object)

    Chla_Load_Cal = np.zeros(X, dtype=object)
    Chla_Load_StL = np.zeros(X, dtype=object)
    Chla_Load_South = np.zeros(X, dtype=object)

    # Water depth could be stage - B.L. (variable) or a constant average depth (e.g., 2.5 m)

    z = np.zeros(X, dtype=object)
    # B_L = 1  # m
    # z = Stage - B_L  # m

    for i in range(X):
        z[i] = 2.5

    fox_min = calibration_parameter_no[19]
    DO_Cr_n = calibration_parameter_no[20]
    DO_Opt_n = calibration_parameter_no[21]
    a = calibration_parameter_no[22]
    DO_Cr_d = calibration_parameter_no[23]
    DO_Opt_d = calibration_parameter_no[24]

    vc = calibration_parameter_chla[28]  # phytoplankton settling velocity (m/d).

    K_r = calibration_parameter_chla[25]
    Theta_r = 1.06
    K_r_T = K_r * Theta_r ** (temperature - 20)

    YNHChla = calibration_parameter_chla[26]  # 0.05
    # Chla mortality
    k_m = calibration_parameter_chla[27]
    Theta_m = 1.06
    K_m_T = k_m * Theta_m ** (temperature - 20)

    Im = calibration_parameter_chla[31]

    Grazing = 0
    Theta_G = 1.06
    Grazing_T = Grazing * Theta_G ** (temperature - 20)
    # Initial values
    NO_N[0] = nitrogen_oxide_north_observed[0]
    NO_S[0] = nitrogen_oxide_south_observed[0]

    NO_MEAN[0] = (NO_N[0] + NO_S[0]) * 0.5

    Sim_Chla_N[0] = chlorophyll_a_north[0]
    Sim_Chla_S[0] = chlorophyll_a_south[0]
    Sim_Chla[0] = chlorophyll_a[0]

    nitro_model_output = pd.DataFrame(inflows['date'], columns=['date'])
    nitro_model_output['date'] = pd.to_datetime(nitro_model_output['date'])
    print("LOONE Nitrogen Module is Running!")
    for i in range(X - 1):
        # print(Nitro_Model_Output['date'].iloc[i])

        Q_I_M[i], Q_O_M[i], External_NO_M[i], External_Chla_M[i] = _calculate_inflows_and_outflows(i, storage_dev,
                                                                                                   q_i, q_o,
                                                                                                   external_nitrite_nitrate,
                                                                                                   s65e_nitrite_nitrate,
                                                                                                   external_chlorophyll_a,
                                                                                                   s65e_chlorophyll_a)

        Q_N2S[i] = (Q_I_M[i] * 1 + Q_O_M[i] * 0)

        t = np.linspace(1, 2, num=2)

        # Nitrification calcs
        Nit_R_N_T_NO[i], Nit_R_S_T_NO[i] = _calculate_nitrification_rates(i, loone_nchla_fns, nit_denit_opt,
                                                                          total_nitrification_rate, kni0_no, kni_no,
                                                                          ammonium_north_observed,
                                                                          ammonium_south_observed, K_NH_NO,
                                                                          dissolved_oxygen, KDO_NO, fox_min, DO_Cr_n,
                                                                          DO_Opt_n, a, theta_ni, temperature)

        # Denitrification calcs
        Denit_R_N_T_NO[i], Denit_R_S_T_NO[i] = _calculate_denitrification_rates(i, loone_nchla_fns, nit_denit_opt,
                                                                                total_denitrification_rate, kden0_no,
                                                                                kden_no, NO_N, NO_S, K_Nitr_NO,
                                                                                dissolved_oxygen, KDO_NO, DO_Cr_d,
                                                                                DO_Opt_d, theta_deni, temperature)

        _calculate_nitrogen_oxide(i, loone_nchla_fns, temperature, T_opt_NO, T_min_NO, T_max_NO, photo_period, rad,
                                  Kw_NO, Kc_NO, Sim_Chla, z, K1_NO, K2_NO, NO_N, t, external_nitrite_nitrate,
                                  atmospheric_deposition_north, Q_N2S, Nit_R_N_T_NO, Denit_R_N_T_NO,
                                  ammonium_north_observed, volume_north, G_max_NO, f_P_N_NO, f_N_N_NO, K_NH_NO, K_TN_NO,
                                  YNOChla_NO, Sim_Chla_N, S_NO_NO, Area_N, NO_Temp_Adj, atmospheric_deposition_south,
                                  Q_O_M, NO_S, Nit_R_S_T_NO, Denit_R_S_T_NO, ammonium_south_observed, volume_south,
                                  Sim_Chla_S, Area_S, NO_MEAN, f_P_S_NO, f_N_S_NO)

        NO_Load_Cal[i], NO_Load_StL[i], NO_Load_South[i] = _calculate_no_loads(i, NO_S, s77_outflow, s308_outflow,
                                                                               total_regional_outflow_south)

        ##### Chla
        _calculate_chlorophyll_a(i, loone_nchla_fns, temperature, T_opt_Chla, T_min_Chla, T_max_Chla,
                                 nitro_model_output, photo_period, rad, Kw_Chla, Kc_Chla, Sim_Chla, z,
                                 K1_Chla, K2_Chla, External_Chla_M, Q_N2S, Sim_Chla_N, vc, K_m_T, K_r_T, G_max_Chla,
                                 f_P_N_Chla, f_N_N_Chla, volume_north, Q_O_M, Sim_Chla_S, f_P_S_Chla, f_N_S_Chla,
                                 volume_south)

        Chla_Load_Cal[i], Chla_Load_StL[i], Chla_Load_South[i] = _calculate_chla_loads(i, Sim_Chla_S, s77_outflow,
                                                                                       s308_outflow,
                                                                                       total_regional_outflow_south)

    print("Exporting Module Outputs!")

    nitro_model_output, nitro_mod_out_m = _update_nitro_model_output(nitro_model_output, NO_N, NO_S, NO_MEAN, Sim_Chla,
                                                                     Sim_Chla_N, Sim_Chla_S, temperature)

    # Constituent Loads
    constit_loads_df, constit_loads_m = _prepare_constit_loads_df(inflows, NO_Load_Cal, NO_Load_StL, NO_Load_South,
                                                                  Chla_Load_Cal, Chla_Load_StL, Chla_Load_South)

    # Summer month loads
    Smr_Mnth_NOx_StL_arr, Smr_Mnth_Chla_StL_arr, Smr_Mnth_NOx_Cal_arr, Smr_Mnth_Chla_Cal_arr = _calculate_summer_month_loads(constit_loads_m)

    if config['sim_type'] in [0, 1]:
        variables_dict = {
            'Constit_Loads': constit_loads_df,
            'Nitro_Model_Output': nitro_model_output,
            'Constit_Loads_M': constit_loads_m,
            'Nitro_Mod_Out_M': nitro_mod_out_m,
        }
        return_list = list(variables_dict.values())

        for k, v in variables_dict.items():
            file_name = f'{k}_forecast' if forecast_mode else k
            v.to_csv(os.path.join(workspace, f'{file_name}.csv'), index=False)
        return return_list
    else:
        return_list = [Smr_Mnth_NOx_StL_arr, Smr_Mnth_Chla_StL_arr, Smr_Mnth_NOx_Cal_arr, Smr_Mnth_Chla_Cal_arr,
                       nitro_model_output]

    # Algae_Opt_Mnth_NOx_StL = []
    # Algae_Opt_Mnth_NOx_Cal = []
    # Algae_Opt_Mnth_Chla_StL = []
    # Algae_Opt_Mnth_Chla_Cal = []

    # for i in range(len(Constit_Loads_M.index)):
    #     if Nitro_Model_Output['Water Temp'].iloc[i] >= 25:
    #         Algae_Opt_Mnth_NOx_StL.append(Constit_Loads_M['NO_Load_StL'].iloc[i])
    #         Algae_Opt_Mnth_Chla_StL.append(Constit_Loads_M['Chla_Load_StL'].iloc[i])
    #         Algae_Opt_Mnth_NOx_Cal.append(Constit_Loads_M['NO_Load_Cal'].iloc[i])
    #         Algae_Opt_Mnth_Chla_Cal.append(Constit_Loads_M['Chla_Load_Cal'].iloc[i])

    # Algae_Opt_Mnth_NOx_StL_arr = np.asarray(Algae_Opt_Mnth_NOx_StL)
    # Algae_Opt_Mnth_Chla_StL_arr = np.asarray(Algae_Opt_Mnth_Chla_StL)

    # Algae_Opt_Mnth_NOx_Cal_arr = np.asarray(Algae_Opt_Mnth_NOx_Cal)
    # Algae_Opt_Mnth_Chla_Cal_arr = np.asarray(Algae_Opt_Mnth_Chla_Cal)
    # if config['sim_type'] in [0, 1]:
    #     return[Constit_Loads_M,Nitro_Mod_Out_M,Algae_Opt_Mnth_NOx_StL_arr,Algae_Opt_Mnth_Chla_StL_arr,Algae_Opt_Mnth_NOx_Cal_arr,Algae_Opt_Mnth_Chla_Cal_arr]
    # else:
    #     return[Algae_Opt_Mnth_NOx_StL_arr,Algae_Opt_Mnth_Chla_StL_arr,Algae_Opt_Mnth_NOx_Cal_arr,Algae_Opt_Mnth_Chla_Cal_arr,Nitro_Model_Output]


if __name__ == "__main__":
    # Parse Arguments
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "workspace",
        nargs=1,
        help="The path to the working directory.",
    )

    args = argparser.parse_args()
    workspace = args.workspace[0]

    # Run LOONE_WQ
    LOONE_WQ(workspace)



def thc_class_normal_or_above(tributary_hydrologic_condition: float, thc_threshold: float) -> bool:
    """
    Determines if the tributary hydrologic condition (THC) is normal or above.

    Args:
        tributary_hydrologic_condition (float): The tributary hydrologic condition to check.
        thc_threshold (float): The threshold for determining if the condition is normal or above.

    Returns:
        bool: True if the tributary hydrologic condition is normal or above, False otherwise.
    """
    return tributary_hydrologic_condition > thc_threshold


def LStgCorres(
    month: int,
    wsm1: float,
    target_stage: float,
    option: int,
    low_chance: int,
) -> float:
    """
    Calculates the lake stage that is to be used with LowChance_Check.

    Args:
        month (int): The current month of the year (1-12).
        wsm1 (float): A parameter from a water surface model.
        target_stage (float): The target stage value under low chance conditions.
        option (int): The option value being used for Opt_LChance_line (in pre_defined_variables.py).
        low_chance (int): A predefined variable related to the low chance conditions.

    Returns:
        float: The calculated corresponding lake stage.
    """
    if option == 1:
        if 5 < month < 10:
            stage = wsm1
        else:
            stage = wsm1 + 0.5
    elif low_chance == 1 or (5 < month < 10):
        stage = -999
    else:
        stage = target_stage
    return stage


def LowChance_Check(lake_stage: float, corresponding_stage: float) -> bool:
    """
    Determines if the Lake Okeechobee stage is greater than a corresponding stage,
    indicating a low chance of the stage falling below a certain level by June 1st.

    Args:
        lake_stage (float): The current water level of Lake Okeechobee.
        corresponding_stage (float): The reference water level to compare against.

    Returns:
        bool: True if the Lake Okeechobee stage is greater than the corresponding stage, False otherwise.
    """
    return lake_stage > corresponding_stage


def Forecast_D_Sal(opt_sal_fcast: int) -> float:
    """
    Calculates the forecasted daily salinity at a specific location based on the provided option.

    Args:
        opt_sal_fcast (int): The option being used for salinity forecast.

    Returns:
        float: The forecasted daily salinity at the specified location.
    """
    import numpy as np

    forecast = np.nan  # Default to NaN if option is not handled

    if opt_sal_fcast == 1:
        # Add logic for option 1
        pass
    elif opt_sal_fcast == 2:
        # Add logic for option 2
        pass
    elif opt_sal_fcast == 3:
        forecast = np.nan

    return forecast


def n30d_mavg(opt_sal_fcast: int) -> float:
    """
    Calculates the 30-day moving average salinity (psu).

    Args:
        opt_sal_fcast (int): Option for salinity forecast.

    Returns:
        float: The 30-day moving average salinity.
    """
    import numpy as np

    moving_avg_salinity = np.nan

    if opt_sal_fcast == 1:
        # Logic for option 1
        pass
    elif opt_sal_fcast == 2:
        # Logic for option 2
        pass
    elif opt_sal_fcast == 3:
        moving_avg_salinity = np.nan

    return moving_avg_salinity


def n30davgForecast(
    estuary_needs_water: float,
    moving_avg_2weeks: list,
    opt_sal_fcast: int,
    sal_threshold: float,
) -> bool:
    """
    Determines if the 30-day average forecast salinity exceeds the defined salinity threshold within two weeks.

    Args:
        estuary_needs_water (float): The current water needs of the estuary.
        moving_avg_2weeks (list): Two weeks' worth of 30-day moving average salinity values.
        opt_sal_fcast (int): The salinity forecast option being used.
        sal_threshold (float): The predefined salinity threshold for the Caloosahatchee Estuary.

    Returns:
        bool: True if the 30-day average forecast salinity exceeds the threshold within two weeks, False otherwise.
    """
    if opt_sal_fcast == 3:
        return estuary_needs_water
    return max(moving_avg_2weeks) > sal_threshold


def LORS08_bf_rel(S77BS_AP: float, n30davgForecast: bool) -> bool:
    """
    Returns whether or not the LORS08 schedule suggests baseflow release.

    Args:
        S77BS_AP (float): The baseflow release from Lake Okeechobee. ?
        n30davgForecast (bool): whether the 30 day average forecast salinity is greater than the defined CE_SalThreshold psu within 2wks.

    Returns:
        bool: True if the LORS08 schedule suggests baseflow release, False otherwise.
    """
    if S77BS_AP > 0:
        return n30davgForecast

    return False


def LDS_LC6_1(
    late_dry_season: bool, low_chance_check: bool, option: int
) -> bool:
    """
    Calculates the value of late_dry_season_low_chance based on the provided parameters.

    Args:
        late_dry_season (bool): Whether or not it is a late dry season.
        low_chance_check (bool): The result of the low_chance_check function (see above).
        option (int): The value being used for the late_dry_season_option.

    Returns:
        bool: The calculated value of late_dry_season_low_chance.
    """
    if option == 1:
        late_dry_season_low_chance = low_chance_check if late_dry_season else False
    else:
        late_dry_season_low_chance = low_chance_check
    return late_dry_season_low_chance


def S_O(late_dry_season_low_chance: bool, baseflow_release_from_lake_okeechobee: float) -> int:
    """
    Calculates the salinity output based on the provided parameters.

    Args:
        late_dry_season_low_chance (bool): Whether or not there is a low chance of rainfall in the late dry season.
        baseflow_release_from_lake_okeechobee (float): The baseflow release from Lake Okeechobee.

    Returns:
        int: The salinity output (0 or 1).
    """
    if not late_dry_season_low_chance:
        if baseflow_release_from_lake_okeechobee > 0:
            salinity_output = 1
        else:
            salinity_output = 0
    else:
        salinity_output = 0
    return salinity_output


def All_4(lors08_baseflow_release: bool, late_dry_season_low_chance_6_1: bool) -> bool:
    """
    Determines whether all four conditions are met to trigger a baseflow release.

    Args:
        lors08_baseflow_release (bool): Whether or not the LORS08 schedule suggests baseflow release.
        late_dry_season_low_chance_6_1 (bool): Whether or not there is a low chance of rainfall in the late dry season.

    Returns:
        bool: Whether or not all four conditions are met.
    """
    if lors08_baseflow_release:
        all_four_conditions_met = late_dry_season_low_chance_6_1
    else:
        all_four_conditions_met = False
    return all_four_conditions_met


def Sabf(lake_o_schedule_zone: int) -> bool:
    """
    Calculates the baseflow allowed value based on the Lake Okeechobee schedule zone.

    Args:
        lake_o_schedule_zone (int): The Lake Okeechobee schedule zone.

    Returns:
        bool: The baseflow allowed value (True or False).
    """
    return lake_o_schedule_zone > 3


def Swbf(lake_o_schedule_zone: int) -> bool:
    """
    Determines whether the specified Lake Okeechobee schedule zone allows for a specific condition.

    Args:
        lake_o_schedule_zone (int): The schedule zone of Lake Okeechobee.

    Returns:
        bool: True if the schedule zone is 3, otherwise False.
    """
    return lake_o_schedule_zone == 3


def Swbu(lake_o_schedule_zone: int) -> bool:
    """
    Determines whether the specified Lake Okeechobee schedule zone allows for a specific condition.

    Args:
        lake_o_schedule_zone (int): The schedule zone of Lake Okeechobee.

    Returns:
        bool: True if the schedule zone is 2, otherwise False.
    """
    return lake_o_schedule_zone == 2


def All_4andStage(all_four: bool, sabf: bool) -> bool:
    """
    Checks if all four conditions are met and the stage is above the baseflow.

    Args:
        all_four (bool): Whether all four conditions are met.
        sabf (bool): Whether the stage is above the baseflow.

    Returns:
        bool: True if all four conditions are met and the stage is above the baseflow, otherwise False.
    """
    return all_four and sabf


def All_4andStagein(all_four: bool, sabf: bool) -> bool:
    """
    Checks if all four conditions are met and the stage is not above the baseflow.

    Args:
        all_four (bool): Whether all four conditions are met.
        sabf (bool): Whether the stage is above the baseflow.

    Returns:
        bool: True if all four conditions are met and the stage is not above the baseflow, otherwise False.
    """
    return all_four and not sabf


def P_AP_BF_Stg(
    lake_o_schedule_zone: int, s77bs_ap: float
) -> bool:
    """
    Checks if the Lake Okeechobee schedule zone and S77BS AP meet the conditions
    for post-AP baseflow stage.

    Args:
        lake_o_schedule_zone (int): The schedule zone of Lake Okeechobee.
        s77bs_ap (float): The S77BS AP value.

    Returns:
        bool: True if the conditions are met, otherwise False.
    """
    return 3 < lake_o_schedule_zone < 7 and s77bs_ap > 0


def Logic_test_1(
    all_four_and_stage: bool, p_ap_bf_stg: bool, no_ap_above_bf_sb: int
) -> bool:
    """
    Determines whether all four conditions are met and the stage is above the
    baseflow, or if the post-AP baseflow stage conditions are met.

    Args:
        all_four_and_stage (bool): Whether all four conditions are met and the
            stage is above the baseflow.
        p_ap_bf_stg (bool): Whether the post-AP baseflow stage conditions are
            met.
        no_ap_above_bf_sb (int): The option to not allow AP above baseflow
            stage in the S77BS zone.

    Returns:
        bool: True if all four conditions are met and the stage is above the
            baseflow, or if the post-AP baseflow stage conditions are met,
            otherwise False.
    """
    if no_ap_above_bf_sb == 0:
        logic = all_four_and_stage
    else:
        logic = p_ap_bf_stg
    return logic


def Post_Ap_Baseflow(
    logic_test_1: bool, s77bs_ap: float, all_4and_stagein: bool, choose_1: float
) -> float:
    """
    Calculates the post-AP baseflow.

    Args:
        logic_test_1 (bool): Whether the logic test 1 conditions are met.
        s77bs_ap (float): The S77BS AP value.
        all_4and_stagein (bool): Whether all four conditions are met and the stage
            is not above the baseflow.
        choose_1 (float): The value to choose if all four conditions are met and
            the stage is not above the baseflow.

    Returns:
        float: The post-AP baseflow value.
    """
    if logic_test_1:
        post_ap_baseflow = s77bs_ap
    elif all_4and_stagein:
        post_ap_baseflow = min(s77bs_ap, choose_1)
    else:
        post_ap_baseflow = 0
    return post_ap_baseflow


def S77RSplusPreAPS77bsf(
    s77rs_pre_ap_s77_baseflow: float, lake_o_schedule_zone: int,
    pre_defined_variables_opt_ceews_lowsm: int
) -> bool:
    """Determines whether the S77RS plus pre-AP S77 baseflow is active.

    Args:
        s77rs_pre_ap_s77_baseflow (float): The S77RS pre-AP S77 baseflow value.
        lake_o_schedule_zone (int): The Lake Okeechobee schedule zone.
        pre_defined_variables_opt_ceews_lowsm (int): The pre-defined variable
            option for CE EWS LO WSM.

    Returns:
        bool: Whether the S77RS plus pre-AP S77 baseflow is active.
    """
    if pre_defined_variables_opt_ceews_lowsm == 0:
        result = s77rs_pre_ap_s77_baseflow == 0 and lake_o_schedule_zone > 1
    elif s77rs_pre_ap_s77_baseflow == 0:
        result = True
    else:
        result = False
    return result


def AndEstNeedsLakeWater(n30davg_forecast: bool, s77rs_plus_pre_ap_s77bsf: bool) -> bool:
    """
    Determines if both the 30-day average forecast and the S77RS plus pre-AP S77 baseflow conditions are met.

    Args:
        n30davg_forecast (bool): Whether the 30-day average forecast condition is met.
        s77rs_plus_pre_ap_s77bsf (bool): Whether the S77RS plus pre-AP S77 baseflow condition is met.

    Returns:
        bool: True if both conditions are met, otherwise False.
    """
    return n30davg_forecast and s77rs_plus_pre_ap_s77bsf


def AndLowChance61stagelessth11(
    late_dry_season_low_chance: bool, and_est_needs_lake_water: bool,
) -> bool:
    """Determines if both the late dry season low chance condition and the estuary needs lake water condition are met.

    Args:
        late_dry_season_low_chance (bool): Whether the late dry season low chance condition is met.
        and_est_needs_lake_water (bool): Whether the estuary needs lake water condition is met.

    Returns:
        bool: True if both conditions are met, otherwise False.
    """
    return late_dry_season_low_chance and and_est_needs_lake_water


def ATHCnora(
    and_low_chance_61_stage_less_than_h11: bool,
    thc_class_normal_or_above: bool,
    late_dry_season: bool,
    pre_defined_variables_opt_thc_byp_late_ds: int,
) -> bool:
    """
    Determines if the adaptive threshold for hydrologic conditions is not required
    or above.

    Args:
        and_low_chance_61_stage_less_than_h11 (bool): Whether the late dry season
            low chance condition and the stage is less than 11 feet condition are
            met.
        thc_class_normal_or_above (bool): Whether the tributary hydrologic
            condition is normal or above.
        late_dry_season (bool): Whether it is late dry season.
        pre_defined_variables_opt_thc_byp_late_ds (int): The pre-defined variable
            option for THC by-passing late dry season.

    Returns:
        bool: True if the adaptive threshold for hydrologic conditions is not
            required or above, otherwise False.
    """
    if pre_defined_variables_opt_thc_byp_late_ds == 1:
        at_nora = and_low_chance_61_stage_less_than_h11 and (
            thc_class_normal_or_above or late_dry_season
        )
    else:
        at_nora = and_low_chance_61_stage_less_than_h11 and thc_class_normal_or_above
    return at_nora


def Choose_PAPEWS_1(
    wsm_zone: int,
    a_pape_ws_1_cb1: float,
    a_pape_ws_1_cb2: float,
    a_pape_ws_1_cb3: float,
    a_pape_ws_1_cb4: float,
) -> float:
    """
    Choose one of four coefficients based on the WSM zone.

    Args:
        wsm_zone (int): The WSM zone.
        a_pape_ws_1_cb1 (float): Coefficient for zone 1.
        a_pape_ws_1_cb2 (float): Coefficient for zone 2.
        a_pape_ws_1_cb3 (float): Coefficient for zone 3.
        a_pape_ws_1_cb4 (float): Coefficient for zone 4.

    Returns:
        float: The chosen coefficient.
    """
    coefficients = {
        1: a_pape_ws_1_cb1,
        2: a_pape_ws_1_cb2,
        3: a_pape_ws_1_cb3,
        4: a_pape_ws_1_cb4,
    }

    return coefficients.get(wsm_zone, -9999) / 100


def Choose_PAPEWS_2(opt_ce_ews_low_sm: int, cal_est_ews: float) -> float:
    """
    Choose one of four coefficients based on the CE EWS LO WSM option.

    Args:
        opt_ce_ews_low_sm (int): The CE EWS LO WSM option.
        cal_est_ews (float): The calibrated estimate of EWS.

    Returns:
        float: The chosen coefficient.
    """
    coefficient_map = {
        1: 0,
        2: cal_est_ews,
        3: 300,
    }

    return coefficient_map.get(opt_ce_ews_low_sm, -9999)


def Post_AP_EWS(
    athc_nora: bool,
    wsm_zone: int,
    choose_pape_ws_1: float,
    choose_pape_ws_2: float,
    pre_defined_variables_cal_est_ews: float,
) -> float:
    """Calculate the post-AP EWS value.

    Args:
        athc_nora (bool): Whether the adaptive threshold for hydrologic conditions
            is not required or above.
        wsm_zone (int): The WSM zone.
        choose_pape_ws_1 (float): The chosen coefficient for PAPEWS_1.
        choose_pape_ws_2 (float): The chosen coefficient for PAPEWS_2.
        pre_defined_variables_cal_est_ews (float): The pre-defined variable calibrated
            estimate of EWS.

    Returns:
        float: The post-AP EWS value.
    """
    if athc_nora:
        if wsm_zone == 0:
            pape_ws = pre_defined_variables_cal_est_ews
        else:
            pape_ws = (1 - choose_pape_ws_1) * choose_pape_ws_2
    else:
        pape_ws = 0
    return pape_ws



import os
import numpy as np
import pandas as pd
from datetime import datetime
from loone.utils import load_config, leap_year


# This following section calculates the parameter states (trib conds, stage tests, seasonal & multi-seasonal LONINO) and sets a 4-digit code for the branch.  The branch code is used with the Routing sheet to determine release rates.
def WCA_Stages_Cls(workspace: str, TC_LONINO_df: pd.DataFrame | None):
    """
    Reads in the WCA Stage data and WCA3A_REG inputs, sets up a daily date range for the simulation period,
    and generates a WCA Stage dataframe. It also generates a daily date range for one year (2020), assigns
    the 366 values of the predefined WCA3a_REG_Zone (One random year) in the WCA3A_REG dataframe to the entire
    study period, and assigns the values of the predefined WCA-2A Schedule to the entire study period.

    Args:
        workspace(str): Path to the workspace directory.
        TC_LONINO_df(pd.DataFrame): Tributary Conditions and Lake Okeechobee Water Level data.

    Returns:
        pd.DataFrame: A dataframe containing the WCA stage data, and the corresponding flags for the WCA-3A and WCA-2A
        stages being above or below the regulation schedule.
    """
    os.chdir(workspace)
    config = load_config(workspace)

    year, month, day = map(int, config["start_date_entry"])
    startdate = datetime(year, month, day).date()
    year, month, day = map(int, config["end_date_entry"])
    enddate = datetime(year, month, day).date()

    date_rng_5 = pd.date_range(start=startdate, end=enddate, freq="D")

    # Read the WCA Stage data
    WCA_Stages = pd.read_csv(os.path.join(workspace, config["wca_stages_inputs"]))
    WCA_Stages.columns = WCA_Stages.columns.str.strip()
    # Read WCA3A_REG inputs
    # Note that I added a date column in the first column and I copied values of Feb28 to the Feb29!
    WCA3A_REG = pd.read_csv(os.path.join(workspace, "WCA3A_REG_Inputs.csv"))

    # generate WCA Stage dataframe
    WCA_Stages_df = pd.DataFrame(date_rng_5, columns=["Date"])
    WCA_Stages_df["3A-NW"] = WCA_Stages["3A-NW"]
    WCA_Stages_df["2A-17"] = WCA_Stages["2A-17"]
    WCA_Stages_df["3A-3"] = WCA_Stages["3A-3"]
    WCA_Stages_df["3A-4"] = WCA_Stages["3A-4"]
    WCA_Stages_df["3A-28"] = WCA_Stages["3A-28"]
    WCA_Stages_df["3gage Avg"] = (
        WCA_Stages_df["3A-3"] + WCA_Stages_df["3A-4"] + WCA_Stages_df["3A-28"]
    ) / 3
    # Generate a daily date range for one year (2020)
    date_rng_10 = pd.date_range(start="1/1/2020", end="12/31/2020", freq="D")
    WCA3A_REG["Date"] = date_rng_10
    Simperioddays = len(WCA_Stages_df.index)
    Oneyeardays = len(WCA3A_REG.index)

    # This Following Loop assigns the 366 values of the predifined WCA3a_REG_Zone (One random year) in the WCA3A_REG dataframe
    # to the entire study period.
    def Replicate_WCA3(
        year: int, day_num: int
    ) -> float:  # Where x is the percentage value (i.e., 10,20,30,40,50,60)
        """
        Retrieves the value from the WCA3A_REG DataFrame corresponding to the given day and percentage.
        Takes leap years into account.

        Args:
            year (int): The year that day_num is in.
            day_num (int): The day number to get the value for (ranging from 0 to 365).

        Returns:
            float: The value from the WCA3A_REG DataFrame corresponding to the day number and the given percentage.
        """
        leap_day_val = WCA3A_REG[config["wca3a_reg_zone"]].iloc[59]
        if leap_year(year) is True:
            day_num_adj = day_num
        else:
            day_num_adj = day_num + (1 if day_num >= 60 else 0)
        day_value = (
            leap_day_val
            if day_num_adj == 60
            and leap_year(year) is True
            else WCA3A_REG[config["wca3a_reg_zone"]].iloc[day_num_adj - 1]
        )
        return day_value

    Iden_WCA3Areg = np.zeros(Simperioddays)
    for i in range(Simperioddays):
        Iden_WCA3Areg[i] = Replicate_WCA3(
            WCA_Stages_df["Date"].iloc[i].year,
            WCA_Stages_df["Date"].iloc[i].timetuple().tm_yday,
        )
    Iden_WCA3Areg_c = [x for x in Iden_WCA3Areg if ~np.isnan(x)]
    WCA_Stages_df[config["wca3a_reg_zone"]] = Iden_WCA3Areg_c

    # Define WCA-2A Schedule
    def Replicate_WCA2A(
        year: int, day_num: int
    ) -> float:
        """
        Retrieves the value from the WCA3A_REG DataFrame corresponding to the given day and percentage.
        Takes leap years into account.

        Args:
            year (int): The year that day_num is in.
            day_num (int): The day number to get the value for (ranging from 0 to 365).

        Returns:
            float: The value from the WCA3A_REG DataFrame corresponding to the day number.
        """
        leap_day_val = WCA3A_REG["WCA2Areg"].iloc[59]
        if leap_year(year):
            day_num_adj = day_num
        else:
            day_num_adj = day_num + (1 if day_num >= 60 else 0)
        day_value = (
            leap_day_val
            if day_num_adj == 60 and leap_year(year)
            else WCA3A_REG["WCA2Areg"].iloc[day_num_adj - 1]
        )
        return day_value

    WCA_2A_Sched = np.zeros(Simperioddays)
    for i in range(Simperioddays):
        WCA_2A_Sched[i] = Replicate_WCA2A(
            WCA_Stages_df["Date"].iloc[i].year,
            WCA_Stages_df["Date"].iloc[i].timetuple().tm_yday,
        )
    WCA_2A_Sched_c = [x for x in WCA_2A_Sched if ~np.isnan(x)]
    WCA_Stages_df["WCA-2A Schedule"] = WCA_2A_Sched_c
    # The Following Lines Determine the predifined WCA3A_REG Zone + Offset, and determine some TRUE or
    # FALSE Logical Tests in #SFWMMdata4LOOPS Tab in the Spreadsheet LOOPS Model.
    Iden_WCA3Areg_plus_offset = np.zeros(Simperioddays)
    thr_gag_les_sch_pl_off = np.zeros(Simperioddays)
    thrA28_less_min = np.zeros(Simperioddays)
    thrANW_less_min = np.zeros(Simperioddays)
    twoA17_less_min = np.zeros(Simperioddays)
    WCA_3A_Stg_gr_Iden_WCA3Areg = np.zeros(Simperioddays)
    WCA_2A_Stg_gr_Sched = np.zeros(Simperioddays)
    E3Aor2A = np.zeros(Simperioddays)
    BWCAs_le_reg = np.zeros(Simperioddays)
    All_Cells = np.zeros(Simperioddays)
    for i in range(Simperioddays):
        Iden_WCA3Areg_plus_offset[i] = (
            WCA_Stages_df[config["wca3a_reg_zone"]].iloc[i]
            + config["wca3a_offset"]
        )

        # 3gage < Sched+offset?
        if WCA_Stages_df["3gage Avg"].iloc[i] < Iden_WCA3Areg_plus_offset[i]:
            TF = True
        else:
            TF = False
        thr_gag_les_sch_pl_off[i] = TF
        # FIXME (WCA-3A floor elevation = 7.5')
        # 3A-28 < min?
        if WCA_Stages_df["3A-28"].iloc[i] < config["wca328_min"]:
            TF = True
        else:
            TF = False
        thrA28_less_min[i] = TF
        # FIXME (LSEL=10.5'!!!)
        # 3A-NW < min?
        if WCA_Stages_df["3A-NW"].iloc[i] < config["wca3_nw_min"]:
            TF = True
        else:
            TF = False
        thrANW_less_min[i] = TF
        # FIXME (LSEL=11.1')
        # 2A-17 < min?
        if WCA_Stages_df["2A-17"].iloc[i] < config["wca217_min"]:
            TF = True
        else:
            TF = False
        twoA17_less_min[i] = TF
        # WCA-3A Stg>iden
        if WCA_Stages_df["3gage Avg"].iloc[i] >= Iden_WCA3Areg_plus_offset[i]:
            TF = True
        else:
            TF = False
        WCA_3A_Stg_gr_Iden_WCA3Areg[i] = TF
        # WCA-2A Stg>Sched
        if (
            WCA_Stages_df["2A-17"].iloc[i]
            >= WCA_Stages_df["WCA-2A Schedule"].iloc[i]
        ):
            TF = True
        else:
            TF = False
        WCA_2A_Stg_gr_Sched[i] = TF
        # Either 3A or 2A are above
        if (
            WCA_3A_Stg_gr_Iden_WCA3Areg[i] is True
            or WCA_2A_Stg_gr_Sched[i] is True
        ):
            TF = True
        else:
            TF = False
        E3Aor2A[i] = TF
        # (both WCAs<reg)?
        if E3Aor2A[i] is True:
            TF = False
        elif E3Aor2A[i] is False:
            TF = True
        BWCAs_le_reg[i] = TF
    # Are WCA stages too low? (not OK to do WSA)
    # Options for constraining/limiting WSA based on WCA stages
    # 0 = no constraint on WSA from WCA stages
    # 1 = Offset named "WCA3Aoffset" is added to the 9.5-10.5' WCA3A regulation schedule to establish a lower limit to allow WSA to Lake O.
    # WSA is allowed only if WCA-3A stage is above this lower limit.
    # Currently LOOPS uses the WCA-3A 3-gage avg stage from the 41-yr SFWMM simulation of the LORS-08.
    # 2= This option allows multiple WCA sites to limit WSA to Lake O.  Minimum stages named "WCA2A17min", "WCA3ANmin", "WCA3A28min", etc.  Minimum stage values compared to the SFWMM-simulated stages at these sites.  WSA is allowed only if WCA stages at these sites are all above their respective minimums.
    # See worksheet "SFWMMdata4LOOPS" beginning in column BF
    WCA_Stages_df[f'{config["wca3a_reg_zone"]}+offset'] = (
        Iden_WCA3Areg_plus_offset
    )
    WCA_Stages_df["3gage < Sched+offset?"] = thr_gag_les_sch_pl_off
    WCA_Stages_df["3A-28 < min?"] = thrA28_less_min
    WCA_Stages_df["3A-NW < min?"] = thrANW_less_min
    WCA_Stages_df["2A-17 < min?"] = twoA17_less_min
    WCA_Stages_df["WCA-3A Stg>Iden_WCA3Areg"] = WCA_3A_Stg_gr_Iden_WCA3Areg
    WCA_Stages_df["WCA-2A Stg>Sched"] = WCA_2A_Stg_gr_Sched
    WCA_Stages_df["Either 3A or 2A are above"] = E3Aor2A
    WCA_Stages_df["(both WCAs<reg)?"] = BWCAs_le_reg
    # First Cell
    if config["opt_wca_limit_wsa"] + 1 == 1:
        TF = False
    elif config["opt_wca_limit_wsa"] + 1 == 2:
        TF = WCA_Stages_df["3gage < Sched+offset?"].iloc[0]
    else:
        if (
            WCA_Stages_df["3A-28 < min?"].iloc[0]
            or WCA_Stages_df["3A-NW < min?"].iloc[0]
            or WCA_Stages_df["2A-17 < min?"].iloc[0]
        ):
            TF = True
        else:
            TF = False
        All_Cells[0] = TF
    if WCA_Stages_df["(both WCAs<reg)?"].iloc[1] or (
        WCA_Stages_df["3A-28 < min?"].iloc[1]
        or WCA_Stages_df["3A-NW < min?"].iloc[1]
        or WCA_Stages_df["2A-17 < min?"].iloc[1]
    ):
        TF = True
    else:
        TF = False
        All_Cells[1] = TF
    for i in range(2, Simperioddays):
        if (
            WCA_Stages_df["Either 3A or 2A are above"].iloc[i]
            and WCA_Stages_df["3A-28"].iloc[i] >= config["wca328_min"]
            and WCA_Stages_df["3A-NW"].iloc[i] >= config["wca3_nw_min"]
            and WCA_Stages_df["2A-17"].iloc[i] >= config["wca217_min"]
        ):
            TF = False
        else:
            TF = True
        All_Cells[i] = TF
    WCA_Stages_df["Are WCA stages too low?"] = All_Cells
    return WCA_Stages_df



from loone.utils import ap_functions


def THC_Class(
    config,
    i,
    THC_Class_normal_or_above,
    Lake_O_Stage_AP,
    Lake_O_Schedule_Zone,
    LStgCorres,
    LowChance_Check,
    S77RS_AP,
    S77BS_AP,
    S77RS_Pre_AP_S77_Baseflow,
    Forecast_D_Sal,
    n30d_mavg,
    n30davgForecast,
    LORS08_bf_rel,
    LDS_LC6_1,
    S_O,
    All_4,
    Sabf,
    Swbf,
    Swbu,
    All_4andStage,
    All_4andStagein,
    P_AP_BF_Stg,
    Logic_test_1,
    Post_Ap_Baseflow,
    S77RSplusPreAPS77bsf,
    AndEstNeedsLakeWater,
    AndLowChance61stagelessth11,
    ATHCnora,
    Choose_PAPEWS_1,
    Choose_PAPEWS_2,
    Post_AP_EWS,
    Post_AP_Baseflow_EWS_cfs,
    AdapProt_df,
    Stage_LO,
    Zone_Code,
    df_WSMs,
    Targ_Stg_df,
    S77RS,
    S77BS,
    Estuary_needs_water,
    Choose_1,
    WSM_Zone,
):

    THC_Class_normal_or_above[i] = ap_functions.thc_class_normal_or_above(
        AdapProt_df["Tributary Hydrologic Condition"].iloc[i],
        config["thc_threshold"],
    )
    # Lake Okeechobee Stage
    Lake_O_Stage_AP[i] = Stage_LO[i + 1]
    # Lake Okeechobee Zone Code
    Lake_O_Schedule_Zone[i] = Zone_Code[i + 1]
    #
    LStgCorres[i] = ap_functions.LStgCorres(
        AdapProt_df["date"].iloc[i].month,
        df_WSMs["WSM1"].iloc[i + 1],
        Targ_Stg_df[f'{config["low_chance"]}%'].iloc[i],
        config["opt_l_chance_line"],
        config["low_chance"],
    )
    # Lowchance 6/1 stage falls <11'?
    LowChance_Check[i] = ap_functions.LowChance_Check(
        Lake_O_Stage_AP[i], LStgCorres[i]
    )
    # Read LO_Model S77RS
    S77RS_AP[i] = S77RS[i + 2]
    # define Pre-AP S-77 Baseflow
    # FIXME first run only!
    # LO_Model['S77BS'] = np.nan
    # FIXME
    # LO_Model['S77BS'] = S77BS['S77BS']
    S77BS_AP[i] = S77BS[i + 2]
    # Sum S77RS + S77BS
    S77RS_Pre_AP_S77_Baseflow[i] = S77RS_AP[i] + S77BS_AP[i]
    # Forecast daily salinity at Val_I75 if no baseflow or WES(psu), or at Ft. Myers if no baseflow or WES(psu), or at Val_I75  if No S77 contribution to S79.
    if config["opt_sal_fcast"] == 1:
        S = "Forecast Daily Salinity at Val_I75 if No Baseflow or EWS (psu)"
    elif config["opt_sal_fcast"] == 2:
        S = "Forecast Daily Salinity at Ft. Myers if No Baseflow or EWS (psu)"
    else:
        S = "Forecast Daily Salinity at Val_I75 if No S77 contribution to S79"
    # Option 3 does not use the daily "forecast" salinity from this column (L), nor the 30d ma from Col M.
    # Column N contains the logical T/F pertinent to the AP flowchart.
    # For Option 3 details, see the formula in Col N and corresponding forecast computations in sheet CE_Sal_ValI75.
    Forecast_D_Sal[i] = ap_functions.Forecast_D_Sal(config["opt_sal_fcast"])
    # Calculate 30-day moving avg salinity (psu)
    n30d_mavg[i] = ap_functions.n30d_mavg(config["opt_sal_fcast"])
    # Calculate the 30d avg Forecast salinity > the defined CE_SalThreshold psu within 2wks?
    n30davgForecast[i] = ap_functions.n30davgForecast(
        Estuary_needs_water["Estuary Needs Water?"].iloc[i],
        n30d_mavg[i : i + 13],
        config["opt_sal_fcast"],
        config["ce_sal_threshold"],
    )
    # LORS-08 suggests baseflow release from Lake & estuary needs Lake water
    LORS08_bf_rel[i] = ap_functions.LORS08_bf_rel(
        S77BS_AP[i], n30davgForecast[i]
    )
    #
    LDS_LC6_1[i] = ap_functions.LDS_LC6_1(
        AdapProt_df["Late_Dry_Season"].iloc[i],
        LowChance_Check[i],
        config["late_dry_season_option"],
    )
    # Define Switch at Cell O
    S_O[i] = ap_functions.S_O(LDS_LC6_1[i], S77BS_AP[i])
    # DEfine All 4 prior conditions are TRUE:
    All_4[i] = ap_functions.All_4(LORS08_bf_rel[i], LDS_LC6_1[i])
    # Stage above Baseflow SB?
    Sabf[i] = ap_functions.Sabf(Lake_O_Schedule_Zone[i])
    # Stage within Baseflow SB?
    Swbf[i] = ap_functions.Swbf(Lake_O_Schedule_Zone[i])
    # Stage within Bene.Use SB?
    Swbu[i] = ap_functions.Swbu(Lake_O_Schedule_Zone[i])
    # All 4 and Stage > Baseflow SB?
    All_4andStage[i] = ap_functions.All_4andStage(All_4[i], Sabf[i])
    # All 4 and Stage in Baseflow SB?
    All_4andStagein[i] = ap_functions.All_4andStagein(All_4[i], Sabf[i])
    # Pre-AP BF>0 AND stg>BF Subband
    P_AP_BF_Stg[i] = ap_functions.P_AP_BF_Stg(
        Lake_O_Schedule_Zone[i], S77BS_AP[i]
    )
    # Post-AP Baseflow (cfs)
    # Modified formula on 12/7/2021 to allow testing alternative baseflow scenarios.
    # Changed from:
    # =IF(V11,J11,IF(W11,MIN(J11,450),0))
    # to:
    # =IF(V11,J11,IF(W11,MIN(J11,CHOOSE(Opt_AdapProt,450,ActiveSchedule!$G$11)),0))
    # Create a logic test to be used for Post-AP Baseflow (cfs)
    Logic_test_1[i] = ap_functions.Logic_test_1(
        All_4andStage[i], P_AP_BF_Stg[i], config["opt_no_ap_above_bf_sb"]
    )
    # Main Function
    Post_Ap_Baseflow[i] = ap_functions.Post_Ap_Baseflow(
        Logic_test_1[i], S77BS_AP[i], All_4andStagein[i], Choose_1
    )
    # S77RS+PreAPS77bsf=0 and [Stage above LOWSM]?
    # New logic added July 2021 to test restricted delivery of environmental water supply to Caloosahatchee Estuary when Lake O stage is below the WST line (ie, within the LOWSM band).
    # Opt_CEews_LOWSM=0, same logic as original AP.  Test if stage above LOWSM.
    # =(AND(K11=0,F11>1))
    # Opt_CEews_LOWSM<>0, new logic allows Env WS releases even if stage is in LOWSM.
    # For the logic in this column, formulas were changed from:
    # =(AND(K11=0,F11>1))
    # TO
    # =IF(Opt_CEews_LOWSM=0,AND(K11=0,F11>1)),(K11=0))
    # This allows Env WS releases for any stage, but logic in subsequent columns cuts back the ews release based on the input cutback percentages that apply to the target CE env ws release.
    # See column "Post-AP EWS" comment and formulas in Col AD.
    S77RSplusPreAPS77bsf[i] = ap_functions.S77RSplusPreAPS77bsf(
        S77RS_Pre_AP_S77_Baseflow[i],
        Lake_O_Schedule_Zone[i],
        config["opt_ceews_lowsm"],
    )
    # AND Est Needs Lake Water?
    AndEstNeedsLakeWater[i] = ap_functions.AndEstNeedsLakeWater(
        n30davgForecast[i], S77RSplusPreAPS77bsf[i]
    )
    # AND Low chance 6/1 stage <11'?
    AndLowChance61stagelessth11[i] = ap_functions.AndLowChance61stagelessth11(
        LDS_LC6_1[i], AndEstNeedsLakeWater[i]
    )
    # AND THC normal or above?
    ATHCnora[i] = ap_functions.ATHCnora(
        AndLowChance61stagelessth11[i],
        THC_Class_normal_or_above[i],
        AdapProt_df["Late_Dry_Season"].iloc[i],
        config["opt_thc_byp_late_ds"],
    )
    # Post-AP EWS (cfs)
    # Define Choose_PAPEWS_1 and 2
    Choose_PAPEWS_1[i] = ap_functions.Choose_PAPEWS_1(
        WSM_Zone[i + 2],
        config["apcb1"],
        config["apcb2"],
        config["apcb3"],
        config["apcb4"],
    )
    #
    Choose_PAPEWS_2[i] = ap_functions.Choose_PAPEWS_2(
        config["opt_ceews_lowsm"], config["cal_est_ews"]
    )
    #
    Post_AP_EWS[i] = ap_functions.Post_AP_EWS(
        ATHCnora[i],
        WSM_Zone[i + 2],
        Choose_PAPEWS_1[i],
        Choose_PAPEWS_2[i],
        config["cal_est_ews"],
    )
    # Post-AP Baseflow + EWS (cfs)
    Post_AP_Baseflow_EWS_cfs[i] = Post_Ap_Baseflow[i] + Post_AP_EWS[i]
    THC_Out = {}
    THC_Out["Post_AP_EWS"] = Post_AP_EWS[i]
    THC_Out["Post_Ap_Baseflow"] = Post_Ap_Baseflow[i]
    return THC_Out



import os
import pandas as pd
from loone.utils import load_config


class Data:
    """A class that represents the data used for running LOONE."""

    def __init__(self, working_path: str):
        """
        Initializes the Data object.

        Args:
            working_path (str): The working directory path.
        """
        # Load the configuration
        config = load_config(working_path)
        self.data_dir = working_path

        # Read the data from the configuration
        self.read_data(config)

    def read_data(self, config: dict) -> None:
        """
        Reads all the data from the configuration files.

        Args:
            config (dict): The configuration dictionary.
        """
        self.SFWMM_Daily_Outputs = self._read_csv(
            config, "sfwmm_daily_outputs"
        )
        self.WSMs_RSBKs = self._read_csv(config, "wsms_rsbps")
        self.Weekly_dmd = self._read_csv(config, "losa_wkly_dmd")
        self.Wkly_Trib_Cond = self._read_csv(config, "trib_cond_wkly_data")
        self.LONINO_Seas_data = self._read_csv(config, "seasonal_lonino")
        self.LONINO_Mult_Seas_data = self._read_csv(
            config, "multi_seasonal_lonino"
        )
        self.NetInf_Input = self._read_csv(config, "netflows_acft")
        self.SFWMM_W_dmd = self._read_csv(config, "water_dmd")
        self.RF_Vol = self._read_csv(config, "rf_vol")
        self.ET_Vol = self._read_csv(config, "et_vol")
        self.C44_Runoff = self._read_csv(config, "c44ro")
        self.C43RO_Daily = self._read_csv(config, "c43ro")
        self.Sum_Basin_RO = self._read_csv(config, "basin_ro_inputs")
        self.C43RO = self._read_csv(config, "c43ro_monthly")
        self.C44RO = self._read_csv(config, "c44ro_nonthly")
        self.SLTRIB = self._read_csv(config, "sltrib_monthly")
        self.S80_RegRelRates = self._read_csv(
            config, "s80_regulatory_release_rates"
        )
        self.S77_RegRelRates = self._read_csv(
            config, "s77_regulatory_release_rates"
        )
        self.CE_SLE_turns = self._read_csv(config, "ce_sle_turns_inputs")
        self.Pulses = self._read_csv(config, "pulses_inputs")
        self.Targ_Stg_May_1st = self._read_csv(
            config, "may_1st_lake_stage_below_11ft", parse_dates=["Date"]
        )
        self.Targ_Stg_June_1st = self._read_csv(
            config, "june_1st_lake_stage_below_11ft", parse_dates=["Date"]
        )
        self.Estuary_needs_water = self._read_csv(
            config, "estuary_needs_water_input"
        )
        self.EAA_MIA_RUNOFF = self._read_csv(config, "eaa_mia_ro_inputs")
        self.Storage_dev_df = self._read_csv(config, "storage_deviation")
        self.Cal_Par = self._read_csv(config, "calibration_parameters")

    def _read_csv(self, config: dict, key: str, **kwargs) -> pd.DataFrame:
        """
        Helper function to construct the file path and read the CSV file.

        Args:
            config (dict): The configuration dictionary.
            key (str): The key to look up the file path in the configuration.
            **kwargs: Additional arguments to pass to pd.read_csv.

        Returns:
            pd.DataFrame: The loaded DataFrame.
        """
        file_path = os.path.join(self.data_dir, config[key])
        return pd.read_csv(file_path, **kwargs)



from .data.data import Data
from .loone_q.LOONE_Q import LOONE_Q
from .loone_nut.LOONE_NUT import LOONE_NUT
from .loone_wq.LOONE_WQ import LOONE_WQ

__all__ = ['Data', 'LOONE_Q', 'LOONE_NUT', 'LOONE_WQ']



def Zone_D_Trib(tributary_condition: int, opt_new_tree: int) -> int:
    """
    Determine the zone D tributary condition based on the tributary condition
    and the option for the new tree.

    Args:
        tributary_condition(int):  The tributary condition code.
        opt_new_tree(int): The option for the new tree (0 or 1).

    Returns:
        int:  The zone D tributary condition code (1, 2, 3, or 4).
    """
    if opt_new_tree == 1:
        if tributary_condition in (1, 2):
            zone_d_trib = 1
        elif tributary_condition == 3:
            zone_d_trib = 2
        elif tributary_condition == 4:
            zone_d_trib = 3
        else:
            zone_d_trib = 4
    else:
        if tributary_condition in (1, 2):
            zone_d_trib = 1
        elif tributary_condition == 3:
            zone_d_trib = 2
        elif tributary_condition in (4, 5):
            zone_d_trib = 3
        else:
            zone_d_trib = 4
    return zone_d_trib


def Zone_D_stage(stage_lo: float, wsm_cb: float) -> int:
    """
    Determine the zone D stage based on the current lake stage and the C break point.

    Args:
        stage_lo(float): The current lake stage.
        wsm_cb(float): The stage at which the C break point is located.

    Returns:
        int: The zone D stage code (1 or 2).
    """
    return 2 if stage_lo > wsm_cb else 1


def Zone_D_Seas(lonino_seasonal_classes: int, zone_d_trib: int, opt_new_tree: int) -> int:
    """
    Determine the zone D seasonal class.

    Args:
        lonino_seasonal_classes(int): The LONINO seasonal class.
        zone_d_trib(int): The zone D tributary class.
        opt_new_tree(int): The option for the new tree (0 or 1).

    Returns:
        int: The zone D seasonal class (1 or 2).
    """
    if opt_new_tree == 0:
        zone_d_seas = 2 if lonino_seasonal_classes == 4 else 1
    else:
        if zone_d_trib == 2:
            zone_d_seas = 1 if lonino_seasonal_classes == 1 else 2
        else:
            zone_d_seas = 2 if lonino_seasonal_classes == 4 else 1
    return zone_d_seas


def Zone_D_MSeas(lonino_multiseasonal_classes: int) -> int:
    """
    Determine the zone D multi-seasonal class.

    Args:
        lonino_multiseasonal_classes (int): The LONINO multi-seasonal class.

    Returns:
        int: The zone D multi-seasonal class (1 or 2).
    """
    zone_d_mseas = 1 if lonino_multiseasonal_classes in (1, 2) else 2
    return zone_d_mseas


def Zone_D_Rel_Code(zone_d_branch_code: int, opt_dec_tree: int) -> int:
    """
    Determine the zone D release code based on the zone D branch code and
    the option for the decision tree.

    Args:
        zone_d_branch_code (int): The zone D branch code.
        opt_dec_tree (int): The option for the decision tree (0 or 1).

    Returns:
        int: The zone D release code (-1, 1, or 4).
    """
    if opt_dec_tree == 1:
        if zone_d_branch_code in (
            1111, 1112, 1121, 1122,
            1211, 1212, 1221, 1222,
            2111, 2112, 2121,
            2211, 2212, 2221,
            3111, 3211, 3121, 3221,
            4111, 4121
        ):
            zone_d_rel_code = -1
        elif zone_d_branch_code in (
            2122, 2222, 3112, 3122, 3212, 3222, 4112, 4122, 4211, 4212
        ):
            zone_d_rel_code = 1
        elif zone_d_branch_code in (4221, 4222):
            zone_d_rel_code = 4
        else:
            zone_d_rel_code = float("nan")
    else:
        zone_d_rel_code = 1

    return zone_d_rel_code


def Zone_C_Trib(tributary_condition: int, opt_new_tree: int) -> int:
    """Determine the zone C tributary condition based on the tributary condition and the option for the new tree.

    Args:
        tributary_condition (int): The tributary condition code.
        opt_new_tree (int): The option for the new tree (0 or 1).

    Returns:
        int: The zone C tributary condition code (1, 2, or 3).
    """
    if opt_new_tree == 1:
        if tributary_condition in (1, 2):
            zone_c_trib = 1
        elif tributary_condition in (3, 4):
            zone_c_trib = 2
        else:
            zone_c_trib = 3
    else:
        if tributary_condition in (1, 2, 3):
            zone_c_trib = 1
        elif tributary_condition in (4, 5):
            zone_c_trib = 2
        else:
            zone_c_trib = 3
    return zone_c_trib


def Zone_C_Seas(lonino_seasonal_classes: int, opt_new_tree: int) -> int:
    """Determine the zone C seasonal class based on the LONINO seasonal class and the option for the new tree.

    Args:
        lonino_seasonal_classes (int): The LONINO seasonal class.
        opt_new_tree (int): The option for the new tree (0 or 1).

    Returns:
        int: The zone C seasonal class (1 or 2).
    """
    if opt_new_tree == 1:
        zone_c_seasonal_class = 1 if lonino_seasonal_classes == 1 else 2
    else:
        zone_c_seasonal_class = 1 if lonino_seasonal_classes in (1, 2) else 2
    return zone_c_seasonal_class


def Zone_C_MSeas(lonino_multiseasonal_classes: int) -> int:
    """Determine the zone C multi-seasonal class.

    Args:
        lonino_multiseasonal_classes (int): The LONINO multi-seasonal class.

    Returns:
        int: The zone C multi-seasonal class (1 or 2).
    """
    return 1 if lonino_multiseasonal_classes == 1 else 2


def Zone_C_MetFcast(
    zone_c_seas: int, lonino_seas: int, opt_met_fcast: int
) -> int:
    """Determine the zone C meteorological forecast based on the zone C seasonal class
    and the option for the meteorological forecast.

    Args:
        zone_c_seas (int): The zone C seasonal class.
        lonino_seas (int): The LONINO seasonal class.
        opt_met_fcast (int): The option for the meteorological forecast (0 or 1).

    Returns:
        int: The zone C meteorological forecast (1 or 2).
    """
    if opt_met_fcast == 0:
        met_fcast = zone_c_seas
    else:
        met_fcast = 1 if lonino_seas in (1, 2) else 2
    return met_fcast


def Zone_C_Rel_Code(zone_c_branch_code: int, opt_dec_tree: int) -> int:
    """Determine the zone C release code based on the zone C branch code and the option for the decision tree.

    Args:
        zone_c_branch_code (int): The zone C branch code.
        opt_dec_tree (int): The option for the decision tree (0 or 1).

    Returns:
        int: The zone C release code (-1, 3, 4, or 5).
    """
    if opt_dec_tree == 1:
        if zone_c_branch_code in (1111, 1211):
            zone_c_rel_code = -1
        elif zone_c_branch_code in (1112, 1212):
            zone_c_rel_code = 3
        elif zone_c_branch_code in (
            1121,
            1122,
            1221,
            1222,
            2111,
            2112,
            2121,
            2122,
            2211,
            2212,
            2221,
            2222,
            3111,
            3112,
            3121,
            3122,
        ):
            zone_c_rel_code = 4
        elif zone_c_branch_code in (3211, 3212, 3221, 3222):
            zone_c_rel_code = 5
        else:
            zone_c_rel_code = float("nan")
    else:
        zone_c_rel_code = 4
    return zone_c_rel_code


def Zone_B_Trib(tributary_condition, opt_new_tree):
    """Determine the zone B tributary condition based on the tributary condition
    and the option for the new tree.

    Args:
        tributary_condition (int): The tributary condition code.
        opt_new_tree (int): The option for the new tree (0 or 1).

    Returns:
        int: The zone B tributary condition code (1, 2, or 3).
    """
    if opt_new_tree == 1:
        if tributary_condition in (1, 2):
            zone_b_trib = 1
        elif tributary_condition in (3, 4):
            zone_b_trib = 2
        else:
            zone_b_trib = 3
    else:
        if tributary_condition in (1, 2):
            zone_b_trib = 1
        elif tributary_condition == 6:
            zone_b_trib = 3
        else:
            zone_b_trib = 2
    return zone_b_trib


def Zone_B_Stage(stage_lo: float, season: int) -> int:
    """
    Determine the zone B stage based on the current lake stage and the season.

    Args:
        stage_lo (float): The current lake stage.
        season (int): The current season (1, 2, 3, or 4).

    Returns:
        int: The zone B stage code (1 or 2).
    """
    if stage_lo < 17.5:
        if season in (1, 2, 3, 4):
            zone_b_stage = 1
        else:
            zone_b_stage = 2
    else:
        zone_b_stage = 2
    return zone_b_stage


def Zone_B_Seas(lonino_seasonal_classes: int) -> int:
    """
    Determine the zone B seasonal class based on the LONINO seasonal class.

    Args:
        lonino_seasonal_classes (int): The LONINO seasonal class.

    Returns:
        int: The zone B seasonal class (1 or 2).
    """
    if lonino_seasonal_classes in (1, 2):
        zone_b_seasonal_class = 1
    else:
        zone_b_seasonal_class = 2
    return zone_b_seasonal_class


def Zone_B_Rel_Code(branch_code: int, opt_dec_tree: int) -> int:
    """
    Determine the zone B release code based on the branch code and option for the decision tree.

    Args:
        branch_code (int): The zone B branch code.
        opt_dec_tree (int): The option for the decision tree (0 or 1).

    Returns:
        int: The zone B release code (3, 5, 6, or NaN).
    """
    if opt_dec_tree == 1:
        if branch_code in (1111, 1211):
            return 3
        elif branch_code in (
            1112, 1121, 1122, 1212, 1221, 1222,
            2111, 2112, 2121, 2122, 2211, 2212,
            2221, 2222, 3111, 3112, 3121, 3122,
            3211, 3212, 3221, 3222, 1131, 1132,
            1141, 1142, 1231, 1232, 1241, 1242,
            2131, 2132, 2141, 2142, 2231, 2232,
            2241, 2242
        ):
            return 5
        elif branch_code in (
            3131, 3132, 3141, 3142, 3231, 3232,
            3241, 3242
        ):
            return 6
        else:
            return float('nan')
    else:
        return 5
